# -*- coding: utf-8 -*-

"""Unit test for the lsl.sim.vis module."""

import os
import unittest
import numpy

from lsl.sim import vis
from lsl.common import stations as lwa_common


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class simvis_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.sim.vis
    module."""
    
    def setUp(self):
        """Turn off all numpy warnings."""
        
        numpy.seterr(all='ignore')
        
    def test_build_aa_flat(self):
        """Test building a antenna array object with uniform sky response."""
        
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        
        aa = vis.buildSimArray(lwa1, antennas, freqs, ForceFlat=True)
        
    def test_build_aa(self):
        """Test building a antenna array object with realistic sky response."""
        
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        # Check the number of stands
        self.assertEqual(len(aa.ants), len(antennas))
        
        # Check the frequencies comming out
        for fo, fi in zip(aa.get_afreqs(), freqs):
            self.assertAlmostEqual(fo, fi/1e9, 6)
            
    def test_build_data(self):
        """Test building simulated visibility data"""
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Do a check of keys
        keyList = out.keys()
        for key in ['freq', 'isMasked', 'bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            self.assertTrue(key in keyList)
            
        # Do a check of frequencies
        for fa, fq in zip(out['freq'], freqs):
            self.assertAlmostEqual(fa, fq, 6)
            
        # Do a check to make sure that the entries with secondary keys have them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(key2 in secondaryKeyList)
                
        # Do a check to make sure that the entries with secondary keys also 
        # have lists in them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(type(out[key][key2]).__name__ == 'list')
                
        #
        # Single-channel test
        #
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.array([30e6,])
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Do a check of keys
        keyList = out.keys()
        for key in ['freq', 'isMasked', 'bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            self.assertTrue(key in keyList)
            
        # Do a check of frequencies
        for fa, fq in zip(out['freq'], freqs):
            self.assertAlmostEqual(fa, fq, 6)
            
        # Do a check to make sure that the entries with secondary keys have them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(key2 in secondaryKeyList)
                
        # Do a check to make sure that the entries with secondary keys also 
        # have lists in them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(type(out[key][key2]).__name__ == 'list')
                
    def test_build_data_res(self):
        """Test building simulated visibility data with resolved sources"""
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs, resolve_src=True)
        
        # Do a check of keys
        keyList = out.keys()
        for key in ['freq', 'isMasked', 'bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            self.assertTrue(key in keyList)
            
        # Do a check of frequencies
        for fa, fq in zip(out['freq'], freqs):
            self.assertAlmostEqual(fa, fq, 6)
            
        # Do a check to make sure that the entries with secondary keys have them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(key2 in secondaryKeyList)
                
        # Do a check to make sure that the entries with secondary keys also 
        # have lists in them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(type(out[key][key2]).__name__ == 'list')
                
        #
        # Single-channel test
        #
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.array([30e6,])
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs, resolve_src=True)
        
        # Do a check of keys
        keyList = out.keys()
        for key in ['freq', 'isMasked', 'bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            self.assertTrue(key in keyList)
            
        # Do a check of frequencies
        for fa, fq in zip(out['freq'], freqs):
            self.assertAlmostEqual(fa, fq, 6)
            
        # Do a check to make sure that the entries with secondary keys have them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(key2 in secondaryKeyList)
                
        # Do a check to make sure that the entries with secondary keys also 
        # have lists in them
        for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            secondaryKeyList = out[key].keys()
            for key2 in ['xx', 'yy', 'xy', 'yx']:
                self.assertTrue(type(out[key][key2]).__name__ == 'list')
                
    def test_scale_data(self):
        """Test that we can scale a data dictionary without error"""
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Scale
        amp = vis.scaleData(out, numpy.ones(len(antennas))*2, numpy.zeros(len(antennas)))
        # Delay
        phs = vis.scaleData(out, numpy.ones(len(antennas)), numpy.ones(len(antennas)))
        
        #
        # Single-channel test
        #
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.array([30e6,])
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Scale
        amp = vis.scaleData(out, numpy.ones(len(antennas))*2, numpy.zeros(len(antennas)))
        # Delay
        phs = vis.scaleData(out, numpy.ones(len(antennas)), numpy.ones(len(antennas)))
        
    def test_shift_data(self):
        """Test that we can shift the uvw coordinates of a data dictionary 
        without error"""
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Shift
        sft = vis.shiftData(out, aa)
        
        #
        # Single-channel test
        #
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.array([30e6,])
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Shift
        sft = vis.shiftData(out, aa)
        
    def test_add_noise(self):
        """Test that we can add baseline noise to a data dictionary without
        error"""
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.arange(30e6, 50e6, 1e6)
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Add in the noise
        na = vis.addBaselineNoise(out, 15e3, 0.061)
        
        #
        # Single-channel test
        #
        
        # Setup
        lwa1 = lwa_common.lwa1
        antennas = lwa1.getAntennas()[0:20]
        freqs = numpy.array([30e6,])
        aa = vis.buildSimArray(lwa1, antennas, freqs)
        
        # Build the data dictionary
        out = vis.buildSimData(aa, vis.srcs)
        
        # Add in the noise
        na = vis.addBaselineNoise(out, 15e3, 0.061, bandwidth=1e6)


class  simvis_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.sim.vis units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(simvis_tests)) 


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Example script that reads in TBW data and runs a cross-correlation on it.  
The results are saved in the FITS IDI format.
"""

import os
import sys
import time
import ephem
import numpy
import argparse
from datetime import datetime, timedelta, tzinfo

from lsl import astro
from lsl.reader.ldp import LWA1DataFile
from lsl.common import stations, metabundle
from lsl.correlator import fx as fxc
from lsl.writer import fitsidi
from lsl.common.progress import ProgressBar
from lsl.misc import parser as aph


class UTC(tzinfo):
    """tzinfo object for UTC time."""
    
    def utcoffset(self, dt):
        return timedelta(0)
        
    def tzname(self, dt):
        return "UTC"
        
    def dst(self, dt):
        return timedelta(0)


def processChunk(idf, site, good, filename, LFFT=64, Overlap=1, pols=['xx','yy']):
    """
    Given an lsl.reader.ldp.TBWFile instances and various parameters for the 
    cross-correlation, write cross-correlate the data and save it to a file.
    """
    
    # Get antennas
    antennas = site.getAntennas()
    
    # Get the metadata
    sampleRate = idf.getInfo('sampleRate')
    
    # Create the list of good digitizers and a digitizer to Antenna instance mapping.  
    # These are:
    #  toKeep  -> mapping of digitizer number to array location
    #  mapper -> mapping of Antenna instance to array location
    toKeep = [antennas[i].digitizer-1 for i in good]
    mapper = [antennas[i] for i in good]
    
    # Create a list of unqiue stands to know what style of IDI file to create
    stands = set( [antennas[i].stand.id for i in good] )
    
    wallTime = time.time()
    readT, t, data = idf.read()
    setTime = t
    refTime = t
    
    setDT = datetime.utcfromtimestamp(setTime)
    setDT.replace(tzinfo=UTC())
    print "Working on set #1 (%.3f seconds after set #1 = %s)" % ((setTime-refTime), setDT.strftime("%Y/%m/%d %H:%M:%S.%f"))
    
    # In order for the TBW stuff to actaully run, we need to run in with sub-
    # integrations.  8 sub-integrations (61.2 ms / 8 = 7.7 ms per section) 
    # seems to work ok with a "reasonable" number of channels.
    nSec = 8
    secSize = data.shape[1]/nSec
    
    # Loop over polarizations (there should be only 1)
    for pol in pols:
        print "-> %s" % pol
        try:
            tempVis *= 0    # pylint:disable=undefined-variable
        except NameError:
            pass
            
        # Set up the progress bar so we can keep up with how the sub-integrations 
        # are progressing
        pb = ProgressBar(max=nSec)
        sys.stdout.write(pb.show()+'\r')
        sys.stdout.flush()
        
        # Loop over sub-integrations (set by nSec)
        for k in xrange(nSec):
            blList, freq, vis = fxc.FXMaster(data[toKeep,k*secSize:(k+1)*secSize], mapper, LFFT=LFFT, Overlap=Overlap, IncludeAuto=True, verbose=False, SampleRate=sampleRate, CentralFreq=0.0, Pol=pol, ReturnBaselines=True, GainCorrect=True)
            
            toUse = numpy.where( (freq>=5.0e6) & (freq<=93.0e6) )
            toUse = toUse[0]
            
            try:
                tempVis += vis
            except:
                tempVis = vis
                
            pb.inc(amount=1)
            sys.stdout.write(pb.show()+'\r')
            sys.stdout.flush()
            
        # Average the sub-integrations together
        vis = tempVis / float(nSec)
        
        # Set up the FITS IDI file is we need to
        if pol == pols[0]:
            pol1, pol2 = fxc.pol2pol(pol)
            
            if len(stands) > 255:
                fits = fitsidi.ExtendedIDI(filename, refTime=refTime)
            else:
                fits = fitsidi.IDI(filename, refTime=refTime)
            fits.setStokes(pols)
            fits.setFrequency(freq[toUse])
            fits.setGeometry(site, [a for a in mapper if a.pol == pol1])
            
        # Add the visibilities
        obsTime = astro.unix_to_taimjd(setTime)
        fits.addDataSet(obsTime, readT, blList, vis[:,toUse], pol=pol)
        sys.stdout.write(pb.show()+'\r')
        sys.stdout.write('\n')
        sys.stdout.flush()
    print "->  Cummulative Wall Time: %.3f s (%.3f s per integration)" % ((time.time()-wallTime), (time.time()-wallTime))
    
    fits.write()
    fits.close()
    del(fits)
    del(data)
    del(vis)
    return True


def main(args):
    # Parse command line options
    filename = args.filename

    # Length of the FFT
    LFFT = args.fft_length

    # Setup the LWA station information
    if args.metadata is not None:
        try:
            station = stations.parseSSMIF(args.metadata)
        except ValueError:
            station = metabundle.getStation(args.metadata, ApplySDM=True)
    else:
        station = stations.lwa1
    antennas = station.getAntennas()
    
    idf = LWA1DataFile(filename)
    
    jd = astro.unix_to_utcjd(idf.getInfo('tStart'))
    date = str(ephem.Date(jd - astro.DJD_OFFSET))
    sampleRate = idf.getInfo('sampleRate')
    nInts = idf.getInfo('nFrames') / (30000 * len(antennas) / 2)
    
    # Get valid stands for both polarizations
    goodX = []
    goodY = []
    for i in xrange(len(antennas)):
        ant = antennas[i]
        if ant.getStatus() != 33 and not args.all:
            pass
        else:
            if ant.pol == 0:
                goodX.append(ant)
            else:
                goodY.append(ant)
                
    # Select which polarization to use
    good = []
    for antX in goodX:
        for antY in goodY:
            if antX.stand.id == antY.stand.id:
                good.append(antX.digitizer-1)
                good.append(antY.digitizer-1)
                break
                
    # Report on the valid stands found.  This is a little verbose,
    # but nice to see.
    print "Found %i good stands to use" % (len(good)/2,)
    for i in good:
        print "%3i, %i" % (antennas[i].stand.id, antennas[i].pol)
        
    # Number of frames to read in at once and average
    nFrames = 30000
    nSets = idf.getInfo('nFrames') / (30000*len(antennas)/2)
    
    print "Data type:  %s" % type(idf)
    print "Captures in file: %i (%.3f s)" % (nInts, nInts*30000*400/sampleRate)
    print "=="
    print "Station: %s" % station.name
    print "Date observed: %s" % date
    print "Julian day: %.5f" % jd
    print "Integration Time: %.3f s" % (400*nFrames/sampleRate)
    print "Number of integrations in file: %i" % nSets
    print "=="
    
    leftToDo = 1
    basename = os.path.split(filename)[1]
    basename, ext = os.path.splitext(basename)
    
    fitsFilename = "%s.FITS_1" % basename
    processChunk(idf, station, good, fitsFilename, LFFT=args.fft_length, Overlap=1, pols=args.products)
    
    idf.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='cross-correlate data in a TBW file', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('filename', type=str, 
                        help='filename to correlate')
    parser.add_argument('-m', '--metadata', type=str, 
                        help='name of SSMIF or metadata tarball file to use for mappings')
    parser.add_argument('-l', '--fft-length', type=aph.positive_int, default=16, 
                        help='set FFT length')
    parser.add_argument('-q', '--quiet', dest='verbose', action='store_false', 
                        help='run %(prog)s in silent mode')
    parser.add_argument('-a', '--all', action='store_true', 
                        help='correlated all dipoles regardless of their status')
    pgroup = parser.add_mutually_exclusive_group(required=True)
    pgroup.add_argument('-x', '--xx', dest='products', action='store_const', const=['xx',], 
                        help='compute only the XX polarization product')
    pgroup.add_argument('-y', '--yy', dest='products', action='store_const', const=['yy',], 
                        help='compute only the YY polarization product')
    pgroup.add_argument('-2', '--two-products', dest='products', action='store_const', const=['xx','yy'], 
                        help='compute only the XX and YY polarization products')
    pgroup.add_argument('-4', '--four-products', dest='products', action='store_const', const=['xx','yy','xy','yx'], 
                        help='compute the XX, XY, YX, and YY polarization products')
    args = parser.parse_args()
    main(args)
    



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Module that contains the error classes for the DRX, TBN, and TBW simulated
data writers.  These errors are currently meant to deal with things that can 
invalidate a frame or set of frames.
"""

__version__ = '0.1'
__revision__ = '$Rev: 2386 $'
__all__ = ['baseSimError', 'invalidStand', 'invalidPol', 'invalidBeam', 'invalidTune', 'invalidDataSize',
        'invalidDataType', 'listErrorCodes', 'MinErrorNo', 'MaxErrorNo', 
        '__version__', '__revision__', '__all__']


MinErrorNo = 1
MaxErrorNo = 6


class baseSimError(IOError):
    """
    Base class for problems that can invalidate a data frame.
    """

    def __init__(self, strerror, errno='-1'):
        self.errno = errno
        self.strerror = strerror
        self.filename = None
        self.args = (errno, strerror)
        super(baseSimError, self).__init__()

    def __str__(self):
        return "%s" % self.strerror


class invalidStand(baseSimError):
    """
    Extension to the base class for dealing with frames with out-of-range 
    stand numbers.  The error code is 1.
    """

    def __init__(self):
        self.errno = 1
        self.strerror = 'Stand number is out of range (==0 or >258)'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(invalidStand, self).__init__(self.strerror, errno=self.errno)


class invalidPol(baseSimError):
    """
    Extension to the base class for dealing with frames with out-of-range 
    polarization numbers.  The error code is 2.
    """

    def __init__(self):
        self.errno = 2
        self.strerror = 'Polarization is out of range (!=0 and !=1)'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(invalidPol, self).__init__(self.strerror, errno=self.errno)


class invalidBeam(baseSimError):
    """
    Extension to the base class for dealing with frames with out-of-range 
    DRX beam numbers.  The error code is 3.
    """

    def __init__(self):
        self.errno = 3
        self.strerror = 'Beam is out of range (==0 or >4)'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(invalidBeam, self).__init__(self.strerror, errno=self.errno)


class invalidTune(baseSimError):
    """
    Extension to the base class for dealing with frames with out-of-range 
    DRX tunning numbers.  The error code is 4.
    """

    def __init__(self):
        self.errno = 4
        self.strerror = 'Tunning is out of range (!=1 and !=2)'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(invalidTune, self).__init__(self.strerror, errno=self.errno)


class invalidDataSize(baseSimError):
    """
    Extension to the base class for dealing with frames with data sections that
    have the wrong array size.  The error code is 5.
    """

    def __init__(self):
        self.errno = 5
        self.strerror = 'Data array size is not consistent with frame type'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(invalidDataSize, self).__init__(self.strerror, errno=self.errno)


class invalidDataType(baseSimError):
    """
    Extension to the base class for dealing with frames with data sections that
    have the wrong type (real vs complex).  The error code is 6.
    """

    def __init__(self):
        self.errno = 6
        self.strerror = 'Data array has the wrong general kind'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(invalidDataType, self).__init__(self.strerror, errno=self.errno)


def listErrorCodes(errno=None):
    """
    Function to provide a list of errors defined in this file.  It 
    alternatively takes an error code using the 'errno' keyword and returns its
    description.
    """

    if errno is None:
        for i in range(MinErrorNo, (MaxErrorNo+1)):
            listErrorCodes(errno=i)
    else:
        if errno == 1:
            print("1: Stand number is out of range (==0 or >258)")
        elif errno == 2:
            print("2: Polarization is out of range (!=0 and !=1)")
        elif errno == 3:
            print("3: Beam is out of range (==0 or >4)")
        elif errno == 4:
            print("4: Tunning is out of range (!=1 and !=2)")
        elif errno == 5:
            print("5: Data array size is not consistent with frame type")
        elif errno == 6:
            print("6: Data array has the wrong general kind")
        else:
            print("Unknown error code '%i'" % errno)



# -*- coding: utf-8 -*-

"""Unit test for lsl.catalog module."""

import unittest

from lsl import catalog


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class catalog_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.catalog
    module."""
    
    def test_LWDA(self):
        """Test catalog.LWA_Catalog constructor."""
        
        catalog.CatalogFactory.get_catalog('LWA')
        
    def test_PSR(self):
        """Test catalog.PSR_Catalog contructor."""
        
        catalog.CatalogFactory.get_catalog('PSR')
        
    def test_PKS(self):
        """Test catalog.PKS_Catalog constructor."""
        
        catalog.CatalogFactory.get_catalog('PKS')
        
    def test_PKS90(self):
        """Test catalog.PKS90_Catalog constructor."""
        
        catalog.CatalogFactory.get_catalog('PKS90')
        
    def test_C3C(self):
        """Test catalog.C3C_Catalog constructor."""
        
        catalog.CatalogFactory.get_catalog('3C')
        
    def test_C4C(self):
        """Test catalog.C4C_Catalog constructor."""
        
        catalog.CatalogFactory.get_catalog('4C')
        
    def test_2FGL(self):
        """Test catalog.F2FGL_Catalog constructor."""
        
        catalog.CatalogFactory.get_catalog('2FGL')
        
    def test_get_names(self):
        """Test catalog.CatalogFactory.get_names() method."""
        
        names = catalog.CatalogFactory.get_names()
        self.assertTrue('LWA' in names)
        self.assertTrue('NONSENSE' not in names)
        
    def test_get_catalog(self):
        """Test catalog.CatalogFactory.get_catalog() function."""
        
        cat = catalog.CatalogFactory.get_catalog('LWA')
        self.assertTrue(isinstance(cat, catalog.LWA_Catalog))
        
        self.assertRaises(ValueError, catalog.CatalogFactory.get_catalog, 'NONSENSE')
        
    def test_lookup(self):
        """Test catalog.Catalog.lookup() method."""
        
        cat = catalog.CatalogFactory.get_catalog('LWA')
        
        source = cat.lookup('SgrA')
        self.assertEqual(source.name, 'SgrA')
        self.assertTrue('W24' in source.alias_list)
        self.assertAlmostEqual(source.position.j2000_equ.ra, 266.300, 3)
        self.assertAlmostEqual(source.position.j2000_equ.dec, -28.806, 3)
        
        source = cat.lookup('CasA')
        self.assertEquals(source.name, 'CasA')
        self.assertTrue('NRAO711' in source.alias_list)
        self.assertAlmostEqual(source.position.j2000_equ.ra, 350.850, 3)
        self.assertAlmostEqual(source.position.j2000_equ.dec, 58.815, 3)
        
        source = cat.lookup('CygA')
        self.assertEquals(source.name, 'CygA')
        self.assertTrue('3C405' in source.alias_list)
        self.assertAlmostEqual(source.position.j2000_equ.ra, 299.868, 3)
        self.assertAlmostEqual(source.position.j2000_equ.dec, 40.734, 3) 
        
        source = cat.lookup('NONSENSE')
        self.assertTrue(source is None) 


class catalog_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lwa_user.catalog
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(catalog_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for lsl.reader.ldp module"""

import os
import unittest

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import ldp
from lsl.reader import errors


__revision__ = "$Rev$"
__version__  = "0.1"
__author__    = "Jayce Dowell"


tbwFile = os.path.join(dataPath, 'tests', 'tbw-test.dat')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')
drxFile = os.path.join(dataPath, 'tests', 'drx-test.dat')
drspecFile = os.path.join(dataPath, 'tests', 'drspec-test.dat')

tbfFile = os.path.join(dataPath, 'tests', 'tbf-test.dat')


class ldp_adp_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.reader
    modules."""
    
    ### TBF ###
    
    def test_ldp_tbf(self):
        """Test the LDP interface for a TBF file."""
        
        f = ldp.TBFFile(tbfFile)
        
        # File info
        self.assertEqual(f.getInfo("sampleRate"), 25e3)
        self.assertEqual(f.getInfo("dataBits"), 4)
        self.assertEqual(f.getInfo("nFrames"), 5)
        
        # Read a frame
        frame = f.readFrame()
        
        # Get the remaining frame count
        self.assertEqual(f.getRemainingFrameCount(), f.getInfo("nFrames")-1)
        
        # Reset
        f.reset()
        
        # Close it out
        f.close()
        
    def test_ldp_tbf_nocheck(self):
        """Test the LDP interface for a TBF file."""
        
        f = ldp.TBFFile(tbfFile, ignoreTimeTagErrors=True)
        
        # File info
        self.assertEqual(f.getInfo("sampleRate"), 25e3)
        self.assertEqual(f.getInfo("dataBits"), 4)
        self.assertEqual(f.getInfo("nFrames"), 5)
        
        # Read a frame
        frame = f.readFrame()
        
        # Get the remaining frame count
        self.assertEqual(f.getRemainingFrameCount(), f.getInfo("nFrames")-1)
        
        # Reset
        f.reset()
        
        # Close it out
        f.close()
        
    ### File Type Discovery ###
    
    def test_ldp_discover_tbw(self):
        """Test the LDP LWA1DataFile function of TBW."""
        # TBW
        self.assertRaises(RuntimeError, ldp.LWASVDataFile, tbwFile)
        
    def test_ldp_discover_tbn(self):
        """Test the LDP LWASVDataFile function of TBN."""
        # TBN
        f = ldp.LWASVDataFile(tbnFile)
        self.assertEqual(type(f), ldp.TBNFile)
        
    def test_ldp_discover_drx(self):
        """Test the LDP LWASVDataFile function of DRX."""
        # DRX
        f = ldp.LWASVDataFile(drxFile)
        self.assertEqual(type(f), ldp.DRXFile)
        
    def test_ldp_discover_drspec(self):
        """Test the LDP LWASVDataFile function of DR Spectrometer."""
        # DR Spectrometer
        f = ldp.LWASVDataFile(drspecFile)
        self.assertEqual(type(f), ldp.DRSpecFile)
        
    def test_ldp_discover_tbf(self):
        """Test the LDP LWASVDataFile function of TBF."""
        # TBF
        f = ldp.LWASVDataFile(tbfFile)
        self.assertEqual(type(f), ldp.TBFFile)
        
    def tearDown(self):
        """Cleanup"""
        for handler in list(ldp._open_ldp_files.handlers):
            handler.close()


class ldp_adp_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.reader.ldp 
    unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(ldp_adp_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for regressions in the lsl.common.busy module."""

import sys
import time
import unittest
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
    
from lsl.common import busy

__revision__ = "$Rev: 2470 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class busy_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the regressions in LSL."""
    
    def test_default(self):
        """Test the busy indicator, default options."""
        
        bi = busy.BusyIndicator()
        
        sys.stdout = StringIO()
        
        bi.start()
        time.sleep(1)
        bi.stop()
        
        sys.stdout = sys.__stdout__


class busy_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.busy unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(busy_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
A collection of utilities to help convert RF engineering/communications terms
into radio astronomy terms.  This module is geared toward taking antenna 
parameters in dBi or dBd and converting them into meaningful RA gains in
K/Jy and then using those to get a system equivalent flux density.

.. versionadded:: 1.0.3
"""

import math

from lsl.common.constants import c as speedOfLight, kB

__version__ = '0.1'
__revision__ = '$Rev: 2385 $'
__all__ = ['dBd2dBi', 'dBd2dBi', 'dBi2gain', 'dBd2gain', 'gain2dBi', 'gain2dBd', 'calculateSEFD', 
        'calculateEffectiveArea', 'Jy2dBm', 'dBm2Jy', 
        '__version__', '__revision__', '__all__']


def dBd2dBi(dBd):
    """
    Convert an antenna gain in dBb (gain relative to the maximum gain of a
    half-wave dipole) into a gain in dBi (gain relative to an isotropic 
    antenna).
    """
    
    return dBd + 2.14844


def dBi2dBd(dBi):
    """
    Convert and antenna gain dBi (gain relative to an isotropic antenna)
    into a gain in dBd (gain relative to the maximum gain of a half-wave 
    dipole).
    """
    
    return dBi - 2.14844


def dBi2gain(dBi, frequency):
    """
    Convert an antenna gain in dBi (gain relative to an isotropic antenna) 
    at a particular frequency in Hz into a gain in K/Jy.
    """
    
    # 10^(dBi/10) is the ratio of the effective area of the antenna relative 
    # to an isotropic antenna where the area of the isotropic antenna is 
    # given by:  wavelength^2 / 4 / pi
    return speedOfLight**2 / (8.0*math.pi*kB*frequency**2) * 10**(dBi/10.0)


def dBd2gain(dBd, frequency):
    """
    Convert an antenna gain in dBd (gain relative to the maximum gain of a
    half-wave dipole) at a particular frequency in Hz into a gain in K/Jy
    """
    
    dBi = dBd2dBi(dBd)
    return dBi2gain(dBi, frequency)


def gain2dBi(gain, frequency):
    """
    Convert an antenna gain in K/Jy at a particular frequency in Hz into a
    gain in dBi (gain relative to an isotropic antenna).
    """
    
    # Calculate the area ratio
    areaRatio = 8.0*math.pi*kB*frequency**2 / speedOfLight**2 * gain
    
    # Convert to dBi
    return 10.0*math.log10(areaRatio)


def gain2dBd(gain, frequency):
    """
    Convert an antenna gain in K/Jy at a particular frequency in Hz into a
    gain in dBd (gain relative to the maximum gain of a half-wave dipole).
    """
    
    dBi = gain2dBi(gain, frequency)
    return dBi2dBd(dBi)


def calculateSEFD(Tsys, gain=None, area=None, efficiency=None):
    """
    Given a variety of parameters, calculate the system equivalent flux 
    density in Jy for an antenna.  The parameters are:
    * Tsys - system temperature in K - required
    * gain - antenna gain in K/Jy - optional
    * area - antenna collecting area in m^2 - optional
    * efficiency - aperture efficiency - optional
    
    Of the optional parameters either 'gain' needs to be supplied or
    both 'area' and 'efficiency'.
    """
    
    # Have we been supplied a gain or do we need to calculate it?
    if gain is None:
        ## Do we have everything we need to calculate the gain?
        if area is None or efficiency is None:
            raise RuntimeError("Too few parameters supplied to calculate the SEFD")
            
        ## Looks like it
        gain = efficiency*area / 2.0 / kB
        
    # Calculate the SEFD
    SEFD = Tsys / gain
    
    return SEFD


def calculateEffectiveArea(gain):
    """
    Given the gain of an antenna in K/Jy, calculate the effective collecting
    area in square meters.
    """
    
    # Calculate the area
    area = 2.0*kB*gain
    
    return area


def Jy2dBm(flux, bandwidth, gain):
    """
    Convert a flux density in Jy into a received signal strength in dBm 
    under the assumptions of:
    * signal bandwidth in Hz
    * antenna gain in K/Jy
    """
    
    # Antenna parameters
    area = calculateEffectiveArea(gain)
    
    # Power in mW
    P =  flux * 10**-26 * bandwidth * area * 1000.0
    
    # To dBm
    return 10.0*math.log10(P)


def dBm2Jy(dBm, bandwidth, gain):
    """
    Convert a received signal strength in dBm into a flux density in Jy under
    the assumptions of:
    * signal bandwidth in Hz
    * antenna gain in K/Jy
    """
    
    # Antenna parameters
    area = calculateEffectiveArea(gain)
    
    # Power in mW
    P = 10**(dBm/10.0) / 1000.0
    
    # To Jy
    return P / (10**-26 * bandwidth * area)
    


# -*- coding: utf-8 -*

"""
Module to set up path information for the LWA Software Library.  Two 
variables are defined by this module:
  * module
  * data

module
    specifies the absolute path to the module

data
    the absolute path to the data directory where data files are stored
"""

import os
import imp

__version__ = '0.1'
__revision__ = '$Rev: 2384 $'
__all__ = ['module', 'data', 'moduleBuild', 'dataBuild', '__version__', '__revision__', '__all__']

modInfo = imp.find_module('lsl')
module = os.path.abspath(modInfo[1])
data = os.path.join(module, 'data')

# If we seem to be in the building directory, make the module and 
# data build paths point to the right place.  This is done so that 
# the testing scripts run via:
#	python setup.py test
# access the right files and tests.
# 
# If we don't seem to be in the 
# source directory, moduleBuild points to module and dataBuild 
# points to data.
currentDir = os.path.abspath(os.getcwd())
if os.path.exists(os.path.join(currentDir, 'setup.py')) and os.path.exists(os.path.join(currentDir, 'lsl')):
    modInfoBuild = imp.find_module('lsl', [currentDir])
    moduleBuild =  os.path.abspath(modInfoBuild[1])
    dataBuild = os.path.join(moduleBuild, 'data')
else:
    moduleBuild = module
    dataBuild = data



# -*- coding: utf-8 -*

"""LWA Software Library
Provided packages:
  * lsl.common
  * lsl.reader
  * lsl.writer
  * lsl.correlator
  * lsl.statistics
  * lsl.sim
  * lsl.imaging
  * lsl.misc

Provided modules:
  * lsl.astro
  * lsl.catalog
  * lsl.skymap
  * lsl.transform

See the individual package descriptions for more information.
"""

import os
import sys
import subprocess
from lsl import version
from lsl.common import paths

__version__ = '0.7'
__revision__ = '$Rev: 2383 $'
__author__ = "Jayce Dowell"


def test():
    """
    Internal test fixture
    """
    
    eggPath = os.path.split(paths.module)[0]
    testPath = os.path.join(eggPath, 'tests', 'test_lsl.py')
    subprocess.check_call([sys.executable, testPath, '-v'])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example script to plot the relative response of an isolated LWA antenna 
as a function of azimuth and elevation."""

import os
import sys
import numpy
import argparse

from scipy.interpolate import interp1d

from lsl.common.paths import data as dataPath
from lsl.misc import parser as aph

import matplotlib.pyplot as plt


def main(args):
    # Build the grid of azimuth and elevations to plot
    az = numpy.zeros((90,360))
    alt = numpy.zeros((90,360))
    for i in range(360):
        az[:,i] = i
    for i in range(90):
        alt[i,:] = i
        
    # Build the figure and the axes
    fig = plt.figure()
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    
    # Get the emperical model of the beam and compute it for the correct frequencies
    i = 0
    beamDict = numpy.load(os.path.join(dataPath, 'lwa1-dipole-emp.npz'))
    for beamCoeff in (beamDict['fitX'], beamDict['fitY']):
        alphaE = numpy.polyval(beamCoeff[0,0,:], args.frequency)
        betaE =  numpy.polyval(beamCoeff[0,1,:], args.frequency)
        gammaE = numpy.polyval(beamCoeff[0,2,:], args.frequency)
        deltaE = numpy.polyval(beamCoeff[0,3,:], args.frequency)
        alphaH = numpy.polyval(beamCoeff[1,0,:], args.frequency)
        betaH =  numpy.polyval(beamCoeff[1,1,:], args.frequency)
        gammaH = numpy.polyval(beamCoeff[1,2,:], args.frequency)
        deltaH = numpy.polyval(beamCoeff[1,3,:], args.frequency)
        if args.verbose:
            print "Beam Coeffs. X: a=%.2f, b=%.2f, g=%.2f, d=%.2f" % (alphaH, betaH, gammaH, deltaH)
            print "Beam Coeffs. Y: a=%.2f, b=%.2f, g=%.2f, d=%.2f" % (alphaE, betaE, gammaE, deltaE)
            
        if args.empirical:
            corrDict = numpy.load(os.path.join(dataPath, 'lwa1-dipole-cor.npz'))
            cFreqs = corrDict['freqs']
            cAlts  = corrDict['alts']
            if corrDict['degrees'].item():
                cAlts *= numpy.pi / 180.0
            cCorrs = corrDict['corrs']
            corrDict.close()
            
            if args.frequency/1e6 < cFreqs.min() or args.frequency/1e6 > cFreqs.max():
                print "WARNING: Input frequency of %.3f MHz is out of range, skipping correction" % (args.frequency/1e6,)
                corrFnc = None
            else:
                fCors = cAlts*0.0
                for j in xrange(fCors.size):
                    ffnc = interp1d(cFreqs, cCorrs[:,j], bounds_error=False)
                    fCors[j] = ffnc(args.frequency/1e6)
                corrFnc = interp1d(cAlts, fCors, bounds_error=False)
                
        else:
            corrFnc = None
            
        def BeamPattern(az, alt, corr=corrFnc):
            zaR = numpy.pi/2 - alt*numpy.pi / 180.0 
            azR = az*numpy.pi / 180.0
            
            c = 1.0
            if corrFnc is not None:
                c = corrFnc(alt*numpy.pi / 180.0)
                c = numpy.where(numpy.isfinite(c), c, 1.0)
                
            pE = (1-(2*zaR/numpy.pi)**alphaE)*numpy.cos(zaR)**betaE + gammaE*(2*zaR/numpy.pi)*numpy.cos(zaR)**deltaE
            pH = (1-(2*zaR/numpy.pi)**alphaH)*numpy.cos(zaR)**betaH + gammaH*(2*zaR/numpy.pi)*numpy.cos(zaR)**deltaH

            return c*numpy.sqrt((pE*numpy.cos(azR))**2 + (pH*numpy.sin(azR))**2)
    
        # Calculate the beam
        pattern = BeamPattern(az, alt)

        if i == 0:
            p = ax1.imshow(pattern, origin='lower', vmin=0, vmax=1)
            ax1.set_title('X pol. @ %.2f MHz' % (args.frequency/1e6))
            ax1.set_xlabel('Azimuth [deg.]')
            ax1.set_ylabel('Elevation [deg.]')
            
            cb = fig.colorbar(p, ax=ax1)
            cb.ax.set_ylabel('Relative Response')
            
        else:
            p = ax2.imshow(pattern, origin='lower', vmin=0, vmax=1)
            ax2.set_title('Y pol. @ %.2f MHz' % (args.frequency/1e6))
            ax2.set_xlabel('Azimuth [deg.]')
            ax2.set_ylabel('Elevation [deg.]')
    
            cb = fig.colorbar(p, ax=ax2)
            cb.ax.set_ylabel('Relative Response')
            
        i += 1
    beamDict.close()
    
    # Display
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='plot the relative dipole response for both polarizations of an isolated LWA antenna at a particular frequency', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('-f', '--frequency', type=aph.frequency, default='50.0', 
                        help='frequency in MHz to generate the dipole model for')
    parser.add_argument('-e', '--empirical', action='store_true', 
                        help='enable empirical corrections to the dipole model (valid from 35 to 80 MHz)')
    parser.add_argument('-v', '--verbose', action='store_true', 
                        help='run %(prog)s in verbose mode')
    args = parser.parse_args()
    main(args)
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Given a DR spectrometer file, plot the time averaged spectra for each 
polarization product."""

import sys
import math
import numpy
import ephem
import argparse

from lsl.reader.ldp import LWA1DataFile
from lsl.astro import unix_to_utcjd, DJD_OFFSET
from lsl.misc import parser as aph

import matplotlib.pyplot as plt


def bestFreqUnits(freq):
    """Given a numpy array of frequencies in Hz, return a new array with the
    frequencies in the best units possible (kHz, MHz, etc.)."""
    
    # Figure out how large the data are
    scale = int(math.log10(freq.max()))
    if scale >= 9:
        divis = 1e9
        units = 'GHz'
    elif scale >= 6:
        divis = 1e6
        units = 'MHz'
    elif scale >= 3:
        divis = 1e3
        units = 'kHz'
    else:
        divis = 1
        units = 'Hz'
        
    # Convert the frequency
    newFreq = freq / divis
    
    # Return units and freq
    return (newFreq, units)


def main(args):
    idf = LWA1DataFile(args.filename)
    
    # Basic file informaiton
    nFramesFile = idf.getInfo('nFrames')
    srate = idf.getInfo('sampleRate')
    beam = idf.getInfo('beam')
    beampols = idf.getInfo('beampols')
    tInt = idf.getInfo('tInt')
    LFFT = idf.getInfo('LFFT')
    products = idf.getInfo('dataProducts')
    
    # Offset in frames for beampols beam/tuning/pol. sets
    args.skip = idf.offset(args.skip)
    
    # Number of frames to integrate over
    maxFrames = 10000
    nFrames = int(args.average / tInt)
    args.average = nFrames * tInt
    
    # Number of remaining chunks
    maxFramesTime = maxFrames*tInt
    nChunks = int(math.ceil(1.0*(nFrames)/maxFrames))
    
    # Date & Central Frequnecy
    beginDate = ephem.Date(unix_to_utcjd(idf.getInfo('tStart')) - DJD_OFFSET)
    centralFreq1 = idf.getInfo('freq1')
    centralFreq2 = idf.getInfo('freq2')
    freq = numpy.fft.fftfreq(LFFT, d=1.0/srate)
    freq = numpy.fft.fftshift(freq)
    
    # File summary
    print "Filename: %s" % args.filename
    print "Date of First Frame: %s" % str(beginDate)
    print "Beam: %i" % beam
    print "Tune/Pols: %i" % beampols
    print "Sample Rate: %i Hz" % srate
    print "Tuning Frequency: %.3f Hz (1); %.3f Hz (2)" % (centralFreq1, centralFreq2)
    print "Frames: %i (%.3f s)" % (nFramesFile, 1.0 * nFramesFile*tInt)
    print "---"
    print "Transform Length: %i channels" % LFFT
    print "Integration Time: %.3f s" % tInt
    print "---"
    print "Offset: %.3f s (%i frames)" % (args.skip, args.skip*srate*beampols/4096)
    print "Integration: %.3f s (%i frames; %i frames per beam/tune/pol)" % (args.average, nFrames, nFrames)
    print "Chunks: %i" % nChunks
    
    # Sanity check
    if args.skip/tInt > nFramesFile:
        raise RuntimeError("Requested offset is greater than file length")
    if nFrames > (nFramesFile - args.skip/tInt):
        raise RuntimeError("Requested integration time+offset is greater than file length")
        
    # Master loop over all of the file chunks
    masterWeight = numpy.zeros((nChunks, 2*len(products), LFFT))
    masterSpectra = numpy.zeros((nChunks, 2*len(products), LFFT))
    for i in xrange(nChunks):
        print "Working on chunk #%i of %i" % (i+1, nChunks)
        
        try:
            readT, t, data = idf.read(args.average/nChunks)
        except Exception, e:
            print "Error: %s" % str(e)
            continue
            
        ## Integrate up the chunck
        data = data.mean(axis=1)
        
        ## Save
        for stand in xrange(data.shape[0]):
            masterSpectra[i,stand,:] = data[stand,:]
            masterWeight[i,stand,:] = int(readT*srate/LFFT)
            
        ## We don't really need the data array anymore, so delete it
        del(data)
        
    # Now that we have read through all of the chunks, perform the final averaging by
    # dividing by all of the chunks
    spec = numpy.squeeze( (masterWeight*masterSpectra).sum(axis=0) / masterWeight.sum(axis=0) )
    
    # Frequencies
    freq1 = freq + centralFreq1
    freq2 = freq + centralFreq2
    
    # The plots:  This is setup for the current configuration of 20 beampols
    fig = plt.figure()
    figsX = int(round(math.sqrt(2*len(products))))
    figsY = 2*len(products) / figsX
    # Put the frequencies in the best units possible
    freq1, units1 = bestFreqUnits(freq1)
    freq2, units2 = bestFreqUnits(freq2)
    
    for i in xrange(masterSpectra.shape[1]):
        if i/len(products) == 0:
            freq = freq1
            units = units1
        else:
            freq = freq2
            units = units2
            
        ax = fig.add_subplot(figsX,figsY,i+1)
        currSpectra = numpy.squeeze( numpy.log10(spec[i,:])*10.0 )
        ax.plot(freq, currSpectra, label='%i (avg)' % (i+1))
        
        # If there is more than one chunk, plot the difference between the global 
        # average and each chunk
        if nChunks > 1 and args.disable_chunks:
            for j in range(nChunks):
                # Some files are padded by zeros at the end and, thus, carry no 
                # weight in the average spectra.  Skip over those.
                if masterWeight[j,i,:].sum() == 0:
                    continue
                    
                # Calculate the difference between the spectra and plot
                subspectra = numpy.squeeze( numpy.log10(masterSpectra[j,i,:])*10.0 )
                diff = subspectra - currSpectra
                ax.plot(freq, diff, label='%i' % j)
                
        ax.set_title('Beam %i, Tune. %i, %s' % (beam, i/len(products), products[i % len(products)]))
        ax.set_xlabel('Frequency [%s]' % units)
        ax.set_ylabel('P.S.D. [dB/RBW]')
        ax.set_xlim([freq.min(), freq.max()])
        ax.legend(loc=0)
        
        print "For beam %i, tune. %i, %s maximum in PSD at %.3f %s" % (beam, i/len(products), products[i % len(products)], freq[numpy.where( spec[i,:] == spec[i,:].max() )][0], units)
        
    print "RBW: %.4f %s" % ((freq[1]-freq[0]), units)
    plt.subplots_adjust(hspace=0.35, wspace=0.30)
    plt.show()
    
    # Save spectra image if requested
    if args.output is not None:
        fig.savefig(args.output)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='read in DR spectrometer files and create a collection of time-averaged spectra', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('filename', type=str, 
                            help='filename to process')
    parser.add_argument('-s', '--skip', type=aph.positive_or_zero_float, default=0.0, 
                        help='skip the specified number of seconds at the beginning of the file')
    parser.add_argument('-a', '--average', type=aph.positive_float, default=10.0, 
                        help='number of seconds of data to average for spectra')
    parser.add_argument('-q', '--quiet', dest='verbose', action='store_false',
                        help='run %(prog)s in silent mode')
    parser.add_argument('-d', '--disable-chunks', action='store_true', 
                        help='disable plotting chunks in addition to the global average')
    parser.add_argument('-o', '--output', type=str, 
                        help='output file name for spectra image')
    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function, division
import sys
if sys.version_info > (3,):
    xrange = range
    from functools import cmp_to_key
    
"""
Module for writing correlator output to a UVFITS file.  The classes and 
functions defined in this module are based heavily off the lwda_fits library.

.. note::
    For arrays with between 256 and 2048 antennas, the baseline packing
    follows the MIRIAD convention.
"""

import os
import gc
import re
import math
import ephem
import numpy
import pyfits
from datetime import datetime

from lsl import astro
from lsl.misc import geodesy
from lsl.common import constants
from lsl.writer.fitsidi import StokesCodes
from lsl.misc.total_sorting import cmp_to_total

__version__ = '0.2'
__revision__ = '$Rev: 2674 $'
__all__ = ['UV', 'StokesCodes', 'NumericStokes', '__version__', '__revision__', '__all__']


UVVersion = (1, 0)


def mergeBaseline(ant1, ant2):
    """
    Merge two stand ID numbers into a single baseline.
    """
    
    if ant1 > 255 or ant2 > 255:
        baseline = ant1*2048 + ant2 + 65536
    else:
        baseline = ant1*256 + ant2
        
    return baseline


def splitBaseline(baseline):
    """
    Given a baseline, split it into it consistent stand ID numbers.
    """
    
    if baseline >= 65536:
        ant1 = int((baseline - 65536) // 2048)
        ant2 = int((baseline - 65536) % 2048)
    else:
        ant1 = int(baseline // 256)
        ant2 = int(baseline % 256)
        
    return ant1,ant2


class UV(object):
    """
    Class for storing visibility data and writing the data, along with array
    geometry, frequency setup, etc., to a UVFITS file that can be read into 
    AIPS via the UVLOD task.
    """
    
    class _Antenna(object):
        """
        Holds information describing the location and properties of an antenna.
        """
        
        def __init__(self, id, x, y, z, bits=8):
            self.id = id
            self.x = x
            self.y = y
            self.z = z
            self.levels = bits
            self.polA = {'Type': 'X', 'Angle': 0.0, 'Cal': [0.0, 0.0]}
            self.polB = {'Type': 'Y', 'Angle': 90.0, 'Cal': [0.0, 0.0]}
            
        def getName(self):
            return "LWA%03i" % self.id
            
    class _Frequency:
        """
        Holds information about the frequency setup used in the file.
        """
        
        def __init__(self, offset, channelWidth, bandwidth):
            self.id = 1
            self.bandFreq = offset
            self.chWidth = channelWidth
            self.totalBW = bandwidth
            self.sideBand = 1
            self.baseBand = 0
            
    @cmp_to_total
    class _UVData(object):
        """
        Represents one UV visibility data set for a given observation time.
        """
    
        def __init__(self, obsTime, intTime, baselines, visibilities, pol=StokesCodes['XX'], source='z'):
            self.obsTime = obsTime
            self.intTime = intTime
            self.baselines = baselines
            self.visibilities = visibilities
            self.pol = pol
            self.source = source
            
        def __cmp__(self, y):
            """
            Function to sort the self.data list in order of time and then 
            polarization code.
            """
            
            sID = (self.obsTime, abs(self.pol))
            yID = (y.obsTime,    abs(y.pol)   )
            
            if sID > yID:
                return 1
            elif sID < yID:
                return -1
            else:
                return 0
                
        def time(self):
            return self.obsTime
            
        def getUVW(self, HA, dec, obs):
            Nbase = len(self.baselines)
            uvw = numpy.zeros((Nbase,3), dtype=numpy.float32)
            
            # Phase center coordinates
            # Convert numbers to radians and, for HA, hours to degrees
            HA2 = HA * 15.0 * numpy.pi/180
            dec2 = dec * numpy.pi/180
            lat2 = obs.lat
            
            # Coordinate transformation matrices
            trans1 = numpy.matrix([[0, -numpy.sin(lat2), numpy.cos(lat2)],
                            [1,  0,               0],
                            [0,  numpy.cos(lat2), numpy.sin(lat2)]])
            trans2 = numpy.matrix([[ numpy.sin(HA2),                  numpy.cos(HA2),                 0],
                            [-numpy.sin(dec2)*numpy.cos(HA2),  numpy.sin(dec2)*numpy.sin(HA2), numpy.cos(dec2)],
                            [ numpy.cos(dec2)*numpy.cos(HA2), -numpy.cos(dec2)*numpy.sin(HA2), numpy.sin(dec2)]])
                    
            for i,(a1,a2) in enumerate(self.baselines):
                # Go from a east, north, up coordinate system to a celestial equation, 
                # east, north celestial pole system
                xyzPrime = a1.stand - a2.stand
                xyz = trans1*numpy.matrix([[xyzPrime[0]],[xyzPrime[1]],[xyzPrime[2]]])
                
                # Go from CE, east, NCP to u, v, w
                temp = trans2*xyz
                uvw[i,:] = numpy.squeeze(temp) / constants.c
                
            return uvw
                
        def argsort(self, mapper=None, shift=16):
            packed = []
            for a1,a2 in self.baselines:
                if mapper is None:
                    s1, s2 = a1.stand.id, a2.stand.id
                else:
                    s1, s2 = mapper[a1.stand.id], mapper[a2.stand.id]
                packed.append( mergeBaseline(s1, s2) )
            packed = numpy.array(packed, dtype=numpy.int32)
            
            return numpy.argsort(packed)
            
    def parseRefTime(self, refTime):
        """
        Given a time as either a integer, float, string, or datetime object, 
        convert it to a string in the formation 'YYYY-MM-DDTHH:MM:SS'.
        """
        
        # Valid time string (modulo the 'T')
        timeRE = re.compile(r'\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?')
        
        if type(refTime) in (int, long, float):
            refDateTime = datetime.utcfromtimestamp(refTime)
            refTime = refDateTime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type(refTime) == datetime:
            refTime = refTime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type(refTime) == str:
            # Make sure that the string times are of the correct format
            if re.match(timeRE, refTime) is None:
                raise RuntimeError("Malformed date/time provided: %s" % refTime)
            else:
                refTime = refTime.replace(' ', 'T', 1)
        else:
            raise RuntimeError("Unknown time format provided.")
            
        return refTime
        
    def refTime2AstroDate(self):
        """
        Convert a reference time string to an :class:`lsl.astro.date` object.
        """
        
        dateStr = self.refTime.replace('T', '-').replace(':', '-').split('-')
        return astro.date(int(dateStr[0]), int(dateStr[1]), int(dateStr[2]), int(dateStr[3]), int(dateStr[4]), float(dateStr[5]))
        
    def __init__(self, filename, refTime=0.0, verbose=False, memmap=None, clobber=False):
        """
        Initialize a new UVFITS object using a filename and a reference time 
        given in seconds since the UNIX 1970 ephem, a python datetime object, or a 
        string in the format of 'YYYY-MM-DDTHH:MM:SS'.
        
        .. versionchanged:: 1.1.2
            Added the 'memmap' and 'clobber' keywords to control if the file
            is memory mapped and whether or not to overwrite an existing file, 
            respectively.
        """
        
        # File-specific information
        self.filename = filename
        self.verbose = verbose
        
        # Observatory-specific information
        self.siteName = 'Unknown'
        
        # Observation-specific information
        self.refTime = self.parseRefTime(refTime)
        self.nAnt = 0
        self.nChan = 0
        self.nStokes = 0
        self.refVal = 0
        self.refPix = 0
        self.channelWidth = 0
        
        # Parameters that store the meta-data and data
        self.array = []
        self.freq = []
        self.stokes = []
        self.data = []
        
        # Open the file and get going
        if os.path.exists(filename):
            if clobber:
                os.unlink(filename)
            else:
                raise IOError("File '%s' already exists" % filename)
        self.FITS = pyfits.open(filename, mode='append', memmap=memmap)
        
    def setStokes(self, polList):
        """
        Given a list of Stokes parameters, update the object's parameters.
        """
        
        for pol in polList:
            if type(pol) == str:
                numericPol = StokesCodes[pol.upper()]
            else:
                numericPol = pol
                
            if numericPol not in self.stokes:
                self.stokes.append(numericPol)
                
        # Sort into order of 'XX', 'YY', 'XY', and 'YX' or 'I', 'Q', 'U', and 'V'
        self.stokes.sort()
        if self.stokes[0] < 0:
            self.stokes.reverse()
            
        self.nStokes = len(self.stokes)
        
    def setFrequency(self, freq):
        """
        Given a numpy array of frequencies, set the relevant common observation
        parameters and add an entry to the self.freq list.
        """
        
        self.nChan = len(freq)
        self.refVal = freq[0]
        self.refPix = 1
        self.channelWidth = numpy.abs(freq[1] - freq[0])
        totalWidth = numpy.abs(freq[-1] - freq[0])
        
        freqSetup = self._Frequency(0.0, self.channelWidth, totalWidth)
        self.freq.append(freqSetup)
        
    def setGeometry(self, site, antennas, bits=8):
        """
        Given a station and an array of stands, set the relevant common observation
        parameters and add entries to the self.array list.
        
        .. versionchanged:: 0.4.0
            Switched over to passing in Antenna instances generated by the
            :mod:`lsl.common.stations` module instead of a list of stand ID
            numbers.
        """
        
        # Make sure that we have been passed 2047 or fewer stands
        if len(antennas) > 2047:
            raise RuntimeError("UVFITS supports up to 2047 antennas only, given %i" % len(antennas))
            
        # Update the observatory-specific information
        self.siteName = site.name
        
        stands = []
        for ant in antennas:
            stands.append(ant.stand.id)
        stands = numpy.array(stands)
        
        arrayX, arrayY, arrayZ = site.getGeocentricLocation()
        
        xyz = numpy.zeros((len(stands),3))
        i = 0
        for ant in antennas:
            xyz[i,0] = ant.stand.x
            xyz[i,1] = ant.stand.y
            xyz[i,2] = ant.stand.z
            i += 1
            
        # Create the stand mapper
        mapper = {}
        if stands.max() > 2047:
            enableMapper = True
        else:
            enableMapper = False
            
        ants = []
        topo2eci = site.getECITransform()
        for i in xrange(len(stands)):
            eci = numpy.dot(topo2eci, xyz[i,:])
            ants.append( self._Antenna(stands[i], eci[0], eci[1], eci[2], bits=bits) )
            if enableMapper:
                mapper[stands[i]] = i+1
            else:
                mapper[stands[i]] = stands[i]
                
        # If the mapper has been enabled, tell the user about it
        if enableMapper and self.verbose:
            print("UVFITS: stand ID mapping enabled")
            for key, value in mapper.iteritems():
                print("UVFITS:  stand #%i -> mapped #%i" % (key, value))
                
        self.nAnt = len(ants)
        self.array.append( {'center': [arrayX, arrayY, arrayZ], 'ants': ants, 'mapper': mapper, 'enableMapper': enableMapper, 'inputAnts': antennas} )
        
    def addComment(self, comment):
        """
        Add a comment to data.
        
        .. versionadded:: 1.2.4
        """
        
        try:
            self._comments.append( comment )
        except AttributeError:
            self._comments = [comment,]
            
    def addHistory(self, history):
        """
        Add a history entry to the data.
        
        .. versionadded:: 1.2.4
        """
        
        try:
            self._history.append( history )
        except AttributeError:
            self._history = [history,]
            
    def addDataSet(self, obsTime, intTime, baselines, visibilities, pol='XX', source='z'):
        """
        Create a UVData object to store a collection of visibilities.
        
        .. versionchanged:: 0.4.0
            Switched over to passing in Antenna instances generated by the
            :mod:`lsl.common.stations` module instead of a list of stand ID
            as part of the baselines.
        """
        
        if type(pol) == str:
            numericPol = StokesCodes[pol.upper()]
        else:
            numericPol = pol
            
        self.data.append( self._UVData(obsTime, intTime, baselines, visibilities, pol=numericPol, source=source) )
        
    def write(self):
        """
        Fill in the FITS-IDI file will all of the tables in the 
        correct order.
        """
        
        # Validate
        if self.nStokes == 0:
            raise RuntimeError("No polarization setups defined")
        if len(self.freq) == 0:
            raise RuntimeError("No frequency setups defined")
        if self.nAnt == 0:
            raise RuntimeError("No array geometry defined")
        if len(self.data) == 0:
            raise RuntimeError("No visibility data defined")
            
        # Sort the data set
        self.data.sort()
        
        self._writePrimary()
        self._writeGeometry()
        
        # Clear out the data section
        del(self.data[:])
        gc.collect()
        
    def close(self):
        """
        Close out the file.
        """
        
        self.FITS.flush()
        self.FITS.close()
        
    def _addCommonKeywords(self, hdr, name, revision):
        """
        Added keywords common to all table headers.
        """
        
        hdr['EXTNAME'] = (name, 'UVFITS table name')
        hdr['EXTVER'] = (1, 'table instance number') 
        hdr['TABREV'] = (revision, 'table format revision number')
        
        date = self.refTime.split('-')
        name = "ZA%s%s%s" % (date[0][2:], date[1], date[2])
        hdr['OBSCODE'] = (name, 'zenith all-sky image')
        
        hdr['ARRNAM'] = self.siteName
        hdr['RDATE'] = (self.refTime, 'file data reference date')
        
    def _writePrimary(self):
        """
        Write the primary HDU to file.
        """
        
        self._writeGeometry(dummy=True)
        hrz = astro.hrz_posn(0, 90)
        (arrPos, ag) = self.readArrayGeometry(dummy=True)
        (mapper, inverseMapper) = self.readArrayMapper(dummy=True)
        ids = ag.keys()
        
        obs = ephem.Observer()
        obs.lat = arrPos.lat * numpy.pi/180
        obs.lon = arrPos.lng * numpy.pi/180
        obs.elev = arrPos.elv * numpy.pi/180
        obs.pressure = 0
        
        first = True
        mList = []
        uList = []
        vList = []
        wList = []
        dateList = []
        blineList = []
        for dataSet in self.data:
            # Sort the data by packed baseline
            try:
                order
            except NameError:
                order = dataSet.argsort(mapper=mapper)
                
            # Deal with defininig the values of the new data set
            if dataSet.pol == self.stokes[0]:
                ## Figure out the new date/time for the observation
                utc = astro.taimjd_to_utcjd(dataSet.obsTime)
                date = astro.get_date(utc)
                date.hours = 0
                date.minutes = 0
                date.seconds = 0
                utc0 = date.to_jd()
                
                ## Update the observer so we can figure out where the source is
                obs.date = utc - astro.DJD_OFFSET
                if dataSet.source == 'z':
                    ### Zenith pointings
                    equ = astro.equ_posn( obs.sidereal_time()*180/numpy.pi, obs.lat*180/numpy.pi )
                    
                    ### format 'source' name based on local sidereal time
                    raHms = astro.deg_to_hms(equ.ra)
                    (tsecs, secs) = math.modf(raHms.seconds)
                    name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), int(tsecs * 10.0))
                else:
                    ### Real-live sources (ephem.Body instances)
                    name = dataSet.source.name
                    
                ## Compute the uvw coordinates of all baselines
                if dataSet.source == 'z':
                    RA = obs.sidereal_time()
                    HA = 0.0
                    dec = equ.dec
                else:
                    RA = dataSet.source.ra * 180/numpy.pi
                    HA = obs.sidereal_time() - dataSet.source.ra
                    dec = dataSet.source.dec * 180/numpy.pi
                    
                if first is True:
                    sourceRA, sourceDec = RA, dec
                    first = False
                    
                uvwCoords = dataSet.getUVW(HA, dec, obs)
                
                ## Populate the metadata
                ### Add in the new baselines
                try:
                    blineList.extend( baselineMapped )
                except NameError:
                    baselineMapped = []
                    for o in order:
                        antenna1, antenna2 = dataSet.baselines[o]
                        if mapper is None:
                            stand1, stand2 = antenna1.stand.id, antenna2.stand.id
                        else:
                            stand1, stand2 = mapper[antenna1.stand.id], mapper[antenna2.stand.id]
                        baselineMapped.append( mergeBaseline(stand1, stand2) ) 
                    blineList.extend( baselineMapped )
                    
                ### Add in the new u, v, and w coordinates
                uList.extend( uvwCoords[order,0] )
                vList.extend( uvwCoords[order,1] )
                wList.extend( uvwCoords[order,2] )
                
                ### Add in the new date/time
                dateList.extend( [utc for bl in dataSet.baselines] )
                
                ### Zero out the visibility data
                try:
                    matrix *= 0.0
                except NameError:
                    matrix = numpy.zeros((len(order), 1, 1, self.nChan, self.nStokes, 2), dtype=numpy.float32)
                    
            # Save the visibility data in the right order
            matrix[:,0,0,:,self.stokes.index(dataSet.pol),0] = dataSet.visibilities[order,:].real
            matrix[:,0,0,:,self.stokes.index(dataSet.pol),1] = dataSet.visibilities[order,:].imag
            
            # Deal with saving the data once all of the polarizations have been added to 'matrix'
            if dataSet.pol == self.stokes[-1]:
                mList.append( matrix*1.0 )
                
        nBaseline = len(blineList)
        
        # Create the UV Data table and update its header
        uv = pyfits.GroupData(numpy.concatenate(mList), parnames=['UU', 'VV', 'WW', 'BASELINE', 'DATE'], 
                            pardata=[numpy.array(uList, dtype=numpy.float32), numpy.array(vList, dtype=numpy.float32), 
                                    numpy.array(wList, dtype=numpy.float32), numpy.array(blineList), 
                                    numpy.array(dateList)], bitpix=-32)
        primary = pyfits.GroupsHDU(uv)
        
        primary.header['EXTEND'] = (True, 'indicates UVFITS file')
        primary.header['GROUPS'] = (True, 'indicates UVFITS file')
        primary.header['OBJECT'] = 'BINARYTB'
        primary.header['TELESCOP'] = self.siteName
        primary.header['INSTRUME'] = self.siteName
        primary.header['OBSERVER'] = ('ZASKY', 'zenith all-sky image')
        primary.header['ORIGIN'] = 'LSL'
        primary.header['CORRELAT'] = ('LWASWC', 'Correlator used')
        primary.header['FXCORVER'] = ('1', 'Correlator version')
        primary.header['LWATYPE'] = ('UV-ZA', 'LWA FITS file type')
        primary.header['LWAMAJV'] = (UVVersion[0], 'LWA UVFITS file format major version')
        primary.header['LWAMINV'] = (UVVersion[1], 'LWA UVFITS file format minor version')
        primary.header['DATE-OBS'] = (self.refTime, 'UVFITS file data collection date')
        ts = str(astro.get_date_from_sys())
        primary.header['DATE-MAP'] = (ts.split()[0], 'UVFITS file creation date')
        
        primary.header['CTYPE2'] = ('COMPLEX', 'axis 2 is COMPLEX axis')
        primary.header['CDELT2'] = 1.0
        primary.header['CRPIX2'] = 1.0
        primary.header['CRVAL2'] = 1.0
        
        primary.header['CTYPE3'] = ('STOKES', 'axis 3 is STOKES axis (polarization)')
        if self.stokes[0] < 0:
            primary.header['CDELT3'] = -1.0
        else:
            primary.header['CDELT3'] = 1.0
        primary.header['CRPIX3'] = 1.0
        primary.header['CRVAL3'] = float(self.stokes[0])
        
        primary.header['CTYPE4'] = ('FREQ', 'axis 4 is FREQ axis (frequency)')
        primary.header['CDELT4'] = self.freq[0].chWidth
        primary.header['CRPIX4'] = self.refPix
        primary.header['CRVAL4'] = self.refVal
        
        primary.header['CTYPE5'] = ('RA', 'axis 5 is RA axis (position of phase center)')
        primary.header['CDELT5'] = 0.0
        primary.header['CRPIX5'] = 1.0
        primary.header['CRVAL5'] = sourceRA
        
        primary.header['CTYPE6'] = ('DEC', 'axis 6 is DEC axis (position of phase center)')
        primary.header['CDELT6'] = 0.0
        primary.header['CRPIX6'] = 1.0
        primary.header['CRVAL6'] = sourceDec
        
        primary.header['TELESCOP'] = self.siteName
        primary.header['OBSERVER'] = 'ZASKY'
        primary.header['SORT'] = ('TB', 'data is sorted in [time,baseline] order')
        
        primary.header['VISSCALE'] = (1.0, 'UV data scale factor')
        
        # Write the comments and history
        try:
            for comment in self._comments:
                primary.header['COMMENT'] = comment
            del self._comments
        except AttributeError:
            pass
        primary.header['COMMENT'] = " FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"
        try:
            for hist in self._history:
                primary.header['HISTORY'] = hist
            del self._history
        except AttributeError:
            pass
            
        self.FITS.append(primary)
        self.FITS.flush()
        
    def _writeGeometry(self, dummy=False):
        """
        Define the 'AIPS AN' table .
        """
        
        i = 0
        names = []
        xyz = numpy.zeros((self.nAnt,3), dtype=numpy.float64)
        for ant in self.array[0]['ants']:
            xyz[i,:] = numpy.array([ant.x, ant.y, ant.z])
            names.append(ant.getName())
            i = i + 1
            
        # Antenna name
        c1 = pyfits.Column(name='ANNAME', format='A8', 
                        array=numpy.array([ant.getName() for ant in self.array[0]['ants']]))
        # Station coordinates in meters
        c2 = pyfits.Column(name='STABXYZ', unit='METERS', format='3D', 
                        array=xyz)
        # Station number
        c3 = pyfits.Column(name='NOSTA', format='1J', 
                        array=numpy.array([self.array[0]['mapper'][ant.id] for ant in self.array[0]['ants']]))
        # Mount type (0 == alt-azimuth)
        c4 = pyfits.Column(name='MNTSTA', format='1J', 
                        array=numpy.zeros((self.nAnt,), dtype=numpy.int32))
        # Axis offset in meters
        c5 = pyfits.Column(name='STAXOF', unit='METERS', format='3E', 
                        array=numpy.zeros((self.nAnt,3), dtype=numpy.float32))
        # Feed A polarization label
        c6 = pyfits.Column(name='POLTYA', format='A1', 
                        array=numpy.array([ant.polA['Type'] for ant in self.array[0]['ants']]))
        # Feed A orientation in degrees
        c7 = pyfits.Column(name='POLAA', format='1E', 
                        array=numpy.array([ant.polA['Angle'] for ant in self.array[0]['ants']], dtype=numpy.float32))
        # Feed A polarization parameters
        c8 = pyfits.Column(name='POLCALA', format='2E', 
                        array=numpy.array([ant.polA['Cal'] for ant in self.array[0]['ants']], dtype=numpy.float32))
        # Feed B polarization label
        c9 = pyfits.Column(name='POLTYB', format='A1', 
                        array=numpy.array([ant.polB['Type'] for ant in self.array[0]['ants']]))
        # Feed B orientation in degrees
        c10 = pyfits.Column(name='POLAB', format='1E', 
                        array=numpy.array([ant.polB['Angle'] for ant in self.array[0]['ants']], dtype=numpy.float32))
        # Feed B polarization parameters
        c11 = pyfits.Column(name='POLCALB', format='2E', 
                        array=numpy.array([ant.polB['Cal'] for ant in self.array[0]['ants']], dtype=numpy.float32))
                        
        # Define the collection of columns
        colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11])
        
        # Create the table and fill in the header
        ag = pyfits.new_table(colDefs)
        self._addCommonKeywords(ag.header, 'AIPS AN', 1)
        
        ag.header['EXTVER'] = (1, 'array ID')
        ag.header['ARRNAM'] = self.siteName
        ag.header['FRAME'] = ('GEOCENTRIC', 'coordinate system')
        ag.header['NUMORB'] = (0, 'number of orbital parameters')
        ag.header['FREQ'] = (self.refVal, 'reference frequency (Hz)')
        ag.header['TIMSYS'] = ('UTC', 'time coordinate system')

        date = self.refTime2AstroDate()
        utc0 = date.to_jd()
        gst0 = astro.get_apparent_sidereal_time(utc0)
        ag.header['GSTIA0'] = (gst0 * 15, 'GAST (deg) at RDATE 0 hours')
        
        utc1 = utc0 + 1
        gst1 = astro.get_apparent_sidereal_time(utc1)
        if gst1 < gst0:
            gst1 += 24.0
        ds = gst1 - gst0
        deg = ds * 15.0      
        ag.header['DEGPDY'] = (360.0 + deg, 'rotation rate of the earth (deg/day)')
        
        refDate = self.refTime2AstroDate()
        refMJD = refDate.to_jd() - astro.MJD_OFFSET
        eop = geodesy.getEOP(refMJD)
        if eop is None:
            eop = geodesy.EOP(mjd=refMJD)
            
        ag.header['UT1UTC'] = (eop.utDiff, 'difference UT1 - UTC for reference date')
        ag.header['IATUTC'] = (astro.leap_secs(utc0), 'TAI - UTC for reference date')
        ag.header['POLARX'] = eop.x
        ag.header['POLARY'] = eop.y
        
        ag.header['ARRAYX'] = (self.array[0]['center'][0], 'array ECI X coordinate (m)')
        ag.header['ARRAYY'] = (self.array[0]['center'][1], 'array ECI Y coordinate (m)')
        ag.header['ARRAYZ'] = (self.array[0]['center'][2], 'array ECI Z coordinate (m)')
        
        ag.header['NOSTAMAP'] = (int(self.array[0]['enableMapper']), 'Mapping enabled for stand numbers')
        
        if dummy:
            self.an = ag
            if self.array[0]['enableMapper']:
                self._writeMapper(dummy=True)
                
        else:
            ag.name = 'AIPS AN'
            self.FITS.append(ag)
            self.FITS.flush()
            
            if self.array[0]['enableMapper']:
                self._writeMapper()
                
    def _writeMapper(self, dummy=False):
        """
        Write a fits table that contains information about mapping stations 
        numbers to actual antenna numbers.  This information can be backed out of
        the names, but this makes the extraction more programmatic.
        """
        
        c1 = pyfits.Column(name='ANNAME', format='A8', 
                        array=numpy.array([ant.getName() for ant in self.array[0]['ants']]))
        c2 = pyfits.Column(name='NOSTA', format='1J', 
                        array=numpy.array([self.array[0]['mapper'][ant.id] for ant in self.array[0]['ants']]))
        c3 = pyfits.Column(name='NOACT', format='1J', 
                        array=numpy.array([ant.id for ant in self.array[0]['ants']]))
                        
        colDefs = pyfits.ColDefs([c1, c2, c3])
        
        # Create the ID mapping table and update its header
        nsm = pyfits.new_table(colDefs)
        self._addCommonKeywords(nsm.header, 'NOSTA_MAPPER', 1)
        
        if dummy:
            self.am = nsm
        else:
            nsm.name = 'NOSTA_MAPPER'
            self.FITS.append(nsm)
            self.FITS.flush()
            
    def readArrayGeometry(self, dummy=False):
        """
        Return a tuple with the array geodetic position and the local 
        positions for all antennas defined in the AIPS AN table.
        """
        
        if dummy:
            try:
                ag = self.an
            except AttributeError:
                raise RuntimeError("Temporary 'AIPS AN' table not found.")
                
        else:
            try:
                ag = self.FITS['AIPS AN']
            except IndexError:
                raise RuntimeError("File does not have an 'AIPS AN' table.")
                
        # Array position
        arrayGeo = astro.rect_posn(ag.header['ARRAYX'], ag.header['ARRAYY'], ag.header['ARRAYZ'])
        arrayGeo = astro.get_geo_from_rect(arrayGeo)
        
        # Antenna positions
        antennaGeo = {}
        antenna = ag.data
        antennaID = antenna.field('NOSTA')
        antennaPos = iter(antenna.field('STABXYZ'))
        for id in antennaID:
            antennaGeo[id] = next(antennaPos)
            
        # Return
        return (arrayGeo, antennaGeo)
        
    def readArrayMapper(self, dummy=False):
        """
        Return a tuple with the array NOSTA mapper and inverse mapper (both
        dictionaries.  If the stand IDs have not been mapped, return None for
        both.
        """
        
        if dummy:
            try:
                nsm = self.am
            except AttributeError:
                return (None, None)
                
        else:
            try:
                nsm = self.FITS['NOSTA_MAPPER']
            except KeyError:
                return (None, None)
                
        # Build the mapper and inverseMapper
        mapper = {}
        inverseMapper = {}
        nosta = nsm.data.field('NOSTA')
        noact = nsm.data.field('NOACT')
        for idMapped, idActual in zip(nosta, noact):
            mapper[idActual] = idMapped
            inverseMapper[idMapped] = idActual
            
        # Return
        return (mapper, inverseMapper)



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Module that contains all of the relevant class to build up a representation 
of a interferometer definition file.  The hierarchy of classes is:
 * Project - class that holds all of the information about the project (including
             the observer) and one or more runs.  Technically, a ID file has 
             only one run but this approach allows for the generation of 
             multiple SD files from a single Project object.
 * Observer - class that hold the observer's name and numeric ID
 * Run - class that holds all of the scans associated and the associated
         correlator setup for a with a particular interferometer run
 * Scan - class that hold information about a particular scan.  It
          includes a variety of attributes that are used to convert human-
          readable inputs to SDF data values.  The scan class is 
          further subclasses into:
           - DRX - class for general DRX scan, with sub-classes:
            * Solar - class for solar tracking
            * Jovian - class for Jovian tracking
    
Most class contain 'validate' attribute functions that can be used to determine if the 
project/run/scan are valid or not given the constraints of
the ADP system.

In addition to providing the means for creating interferometer definition files from 
scratch, this module also includes a simple parser for ID files.

.. versionadded:: 1.2.4
"""

import os
import re
import copy
import math
import pytz
import ephem
from datetime import datetime, timedelta

from lsl.transform import Time
from lsl.astro import utcjd_to_unix, MJD_OFFSET, DJD_OFFSET
from lsl.astro import date as astroDate, get_date as astroGetDate

from lsl.common.mcsADP import mjdmpm2datetime, LWA_MAX_NSTD
from lsl.common.adp import freq2word, word2freq, fC
from lsl.common.stations import LWAStation, getFullStations, lwa1
from lsl.reader.drx import filterCodes as DRXFilters
from lsl.reader.drx import FrameSize as DRXSize
from lsl.common import sdfADP as sdf

from lsl.misc.total_sorting import cmp_to_total


__version__ = '0.1'
__revision__ = '$Rev: 2785 $'
__all__ = ['Observer', 'ProjectOffice', 'Project', 'Run', 'Scan', 'DRX', 'Solar', 'Jovian', 'parseIDF',  'getScanStartStop', 'isValid', '__version__', '__revision__', '__all__']

_UTC = pytz.utc
_DRSUCapacityTB = 10


class Observer(object):
    """Class to hold information about an observer."""
    
    def __init__(self, name, id, first=None, last=None):
        self.name = name
        self.first = first
        self.last = last
        self.id = int(id)

    def joinName(self):
        if self.first != '':
            self.name = ', '.join([self.last, self.first])
        else:
            self.name = self.last
        
    def splitName(self):
        try:
            self.last, self.first = self.name.split(', ', 1)
        except ValueError:
            self.last = self.name
            self.first = ''


class ProjectOffice(object):
    """Class to hold comments from the LWA object office.  This class isn't really 
    needed to create ID files, but it is helpful for parsing ID files."""
    
    def __init__(self, project=None, runs=None, scans=None):
        self.project = project
        if runs is None:
            self.runs = []
        else:
            self.runs = runs
        if scans is None:
            self.scans = []
        else:
            self.scans = scans


class Project(object):
    """
    Class to hold all the information about a specific interferometer run for a 
    project/proposal.
    """
    
    def __init__(self, observer, name, id, runs=None, comments=None, projectOffice=None):
        if not isinstance(observer, Observer):
            raise TypeError("Expected 'observer' to be an Observer")
        self.observer = observer
        self.name = name
        self.id = id
        self.comments = comments
        if runs is None:
            self.runs = []
        else:
            if isinstance(runs, Run):
                runs = [runs,]
            elif isinstance(runs, (list, tuple)):
                for i,sess in enumerate(runs):
                    if not isinstance(sess, Run):
                        raise TypeError("Expected index %i of 'runs' to be an Run" % i)
            else:
                raise TypeError("Expected 'runs' to be either a tuple or list of Run or a Run")
            self.runs = runs
        if projectOffice is None:
            self.projectOffice = ProjectOffice()
        else:
            if not isinstance(projectOffice, ProjectOffice):
                raise TypeError("Expected 'projectOffice' to be a ProjectOffice")
            self.projectOffice = projectOffice
            
    def update(self):
        """Update the various runs that are part of this project."""
        
        for ses in self.runs:
            ses.update()
            
    def validate(self, verbose=False):
        """Examine all of the runs and all of their scans to check
        for validity.  If everything is valid, return True.  Otherwise, return
        False."""
        
        failures = 0
        runCount = 1
        for run in self.runs:
            if verbose:
                print("[%i] Validating run %i" % (os.getpid(), runCount))
            if not run.validate(verbose=verbose):
                failures += 1
                
            runCount += 1
            
        if failures == 0:
            return True
        else:
            return False
            
    def _renderFileSize(self, size):
        """Convert a file size in bytes to a easy-to-use string."""
        
        units = 'B'
        if size >= 1024**4:
            size /= 1024.0**4
            units = 'TB'
        elif size >= 1024**3:
            size /= 1024.0**3
            units = 'GB'
        elif size >= 1024**2:
            size /= 1024.0**2
            units = 'MB'
        elif size >= 1024**1:
            size /= 1024.0**1
            units = 'kB'
            
        return "%.2f %s" % (size, units)
        
    def _renderBandwidth(self, filter, filterCodes):
        """Convert a filter number to an easy-to-use string."""
        
        if filterCodes[filter] > 1e6:
            return "%.3f MHz" % (filterCodes[filter]/1e6,)
        elif filterCodes[filter] > 1e3:
            return "%.3f kHz" % (filterCodes[filter]/1e3,)
        else:
            return "%.3f Hz" % (filterCodes[filter],)
            
    def append(self, newRun):
        """Add a new run to the list of runs."""
        
        if not isinstance(newRun, Run):
            raise TypeError('Expected a Run instance')
        self.runs.append(newRun)
        
    def render(self, run=0, verbose=False):
        """Create a run definition file that corresponds to the specified 
        run.  Returns the ID file's contents as a string."""
        
        if not self.validate(verbose=verbose) :
            raise RuntimeError("Invalid run/scan parameters.  Aborting.")
        if run >= len(self.runs):
            raise IndexError("Invalid run index")
        
        self.runs[run].update()
        self.runs[run].scans.sort()
        for obs in self.runs[run].scans:
            obs.dur = obs.getDuration()
            
        ses = self.runs[run]
        try:
            # Try to pull out the project office comments about the run
            pos = self.projectOffice.runs[run]
        except:
            pos = None
        try:
            # Try to pull out the project office comments about the scans
            poo = self.projectOffice.scans[run]
        except:
            poo = []
        # Enforce that the number of project office scan comments match the
        # actual number of scans
        while (len(ses.scans) - len(poo)) > 0:
            poo.append(None)
            
        # Combine the run comments together in an intelligent fashion
        ## Observer comments
        if ses.ucfuser is not None:
            clean = ''
            if ses.comments:
                clean = sdf._usernameRE.sub('', ses.comments)
            ses.comments = 'ucfuser:%s' % ses.ucfuser
            if len(clean) > 0:
                ses.comments += ';;%s' % clean
        ## Project office comments, including the data return method
        if pos != 'None' and pos is not None:
            pos = 'Requested data return method is %s;;%s' % (ses.dataReturnMethod, pos)
            
        ## PI Information
        output = ""
        output = "%sPI_ID            %s\n" % (output, self.observer.id)
        output = "%sPI_NAME          %s\n" % (output, self.observer.name)
        output = "%s\n" % output
        
        ## Project Information
        output = "%sPROJECT_ID       %s\n" % (output, self.id)
        output = "%sPROJECT_TITLE    %s\n" % (output, self.name)
        output = "%sPROJECT_REMPI    %s\n" % (output, self.comments[:4090] if self.comments else 'None provided')
        output = "%sPROJECT_REMPO    %s\n" % (output, self.projectOffice.project)
        output = "%s\n" % output
        
        ## Run Information
        output = "%sRUN_ID           %s\n" % (output, ses.id)
        output = "%sRUN_TITLE        %s\n" % (output, 'None provided' if ses.name is None else ses.name)
        output = "%sRUN_STATIONS     %s\n" % (output, ','.join([station.id for station in ses.stations]))
        output = "%sRUN_CHANNELS     %i\n" % (output, ses.corr_channels)
        output = "%sRUN_INTTIME      %.3f\n" % (output, ses.corr_inttime)
        output = "%sRUN_BASIS        %s\n" % (output, ses.corr_basis)
        output = "%sRUN_REMPI        %s\n" % (output, ses.comments[:4090] if ses.comments else 'None provided')
        output = "%sRUN_REMPO        %s\n" % (output, "Requested data return method is %s" % ses.dataReturnMethod if pos == 'None' or pos is None else pos[:4090])
        output = "%s\n" % output
                    
        ## Scans
        for i,obs in enumerate(ses.scans):
            obsID = i + 1
            
            output = "%sSCAN_ID          %i\n" % (output, obsID)
            output = "%sSCAN_TARGET      %s\n" % (output, obs.target)
            output = "%sSCAN_INTENT      %s\n" % (output, obs.intent)
            output = "%sSCAN_REMPI       %s\n" % (output, obs.comments[:4090] if obs.comments else 'None provided')
            output = "%sSCAN_REMPO       %s\n" % (output, "Estimated data volume for this scan is %s" % self._renderFileSize(obs.dataVolume) if poo[i] == 'None' or poo[i] == None else poo[i])
            output = "%sSCAN_START_MJD   %i\n" % (output, obs.mjd)
            output = "%sSCAN_START_MPM   %i\n" % (output, obs.mpm)
            output = "%sSCAN_START       %s\n" % (output, obs.start.strftime("%Z %Y/%m/%d %H:%M:%S") if type(obs.start).__name__ == 'datetime' else obs.start)
            output = "%sSCAN_DUR         %i\n" % (output, obs.dur)
            output = "%sSCAN_DUR+        %s\n" % (output, obs.duration)
            output = "%sSCAN_MODE        %s\n" % (output, obs.mode)
            if obs.mode == 'TRK_RADEC':
                output = "%sSCAN_RA          %.9f\n" % (output, obs.ra)
                output = "%sSCAN_DEC         %+.9f\n" % (output, obs.dec)
            output = "%sSCAN_FREQ1       %i\n" % (output, obs.freq1)
            output = "%sSCAN_FREQ1+      %.9f MHz\n" % (output, obs.frequency1/1e6)
            output = "%sSCAN_FREQ2       %i\n" % (output, obs.freq2)
            output = "%sSCAN_FREQ2+      %.9f MHz\n" % (output, obs.frequency2/1e6)
            output = "%sSCAN_BW          %i\n" % (output, obs.filter)
            output = "%sSCAN_BW+         %s\n" % (output, self._renderBandwidth(obs.filter, obs.filterCodes))
            ## Alternate phase centers
            if len(obs.alt_phase_centers) > 0:
                output = "%sSCAN_ALT_N             %i\n" % (output, len(obs.alt_phase_centers))
                for i,phase_center in enumerate(obs.alt_phase_centers):
                    output = "%sSCAN_ALT_TARGET[%i]    %s\n" % (output, i+1, phase_center.target)  
                    output = "%sSCAN_ALT_INTENT[%i]    %s\n" % (output, i+1, phase_center.intent) 
                    output = "%sSCAN_ALT_RA[%i]        %.9f\n" % (output, i+1, phase_center.ra)  
                    output = "%sSCAN_ALT_DEC[%i]       %+.9f\n" % (output, i+1, phase_center.dec)
                    
            ## ASP filter setting
            if obs.aspFlt != -1:
                output = "%sSCAN_ASP_FLT     %i\n" % (output, obs.aspFlt)
            ## DRX gain
            if obs.gain != -1:
                output = "%sSCAN_DRX_GAIN    %i\n" % (output, obs.gain)
            output = "%s\n" % output
            
        return output
        
    def writeto(self, filename, run=0, verbose=False, clobber=False):
        """Create a run definition file that corresponds to the specified 
        run and write it to the provided filename."""
        
        if os.path.exists(filename) and not clobber:
            raise RuntimeError("'%s' already exists" % filename)
            
        output = self.render(run=run, verbose=verbose)
        fh = open(filename, 'w')
        fh.write(output)
        fh.close()
        
    def generateSDFs(self, starting_session_id=1, run=0, verbose=False):
        """Convert the ID file into a collection of `lsl.common.sdfADP.Project` instances
        that can be used to write SD files."""
        
        if not self.validate(verbose=verbose) :
            raise RuntimeError("Invalid run/scan parameters.  Aborting.")
        if run >= len(self.runs):
            raise IndexError("Invalid run index")
        
        self.runs[run].update()
        self.runs[run].scans.sort()
        for obs in self.runs[run].scans:
            obs.dur = obs.getDuration()
            
        ses = self.runs[run]
        try:
            # Try to pull out the project office comments about the run
            pos = self.projectOffice.runs[run]
        except:
            pos = None
        try:
            # Try to pull out the project office comments about the scans
            poo = self.projectOffice.scans[run]
        except:
            poo = []
        # Enforce that the number of project office scan comments match the
        # actual number of scans
        while (len(ses.scans) - len(poo)) > 0:
            poo.append(None)
            
        # Build the SDFs
        ## Setup the common information
        start = mjdmpm2datetime(ses.scans[0].getMJD(), ses.scans[0].getMPM())
        new_observer = sdf.Observer(copy.deepcopy(self.observer.name), copy.deepcopy(self.observer.id))
        ## Go!
        sdfs = []
        for i,station in enumerate(ses.stations):
            ### Session
            session = sdf.Session("%s - %s (%i of %i)" % (ses.name, station.id, i+1, len(ses.stations)), 
                                  starting_session_id, observations=[], station=station)
            session.setDRXBeam(1)
            session.setUCFUsername('eLWA/%s_%s_%s_%04i' % (self.id, start.strftime('%y%m%d'), start.strftime('%H%M'), ses.id))
            session.setDataReturnMethod('UCF')
            
            ## Project Office
            new_projoff = sdf.ProjectOffice(project=copy.deepcopy(self.projectOffice.project), 
                                            sessions=copy.deepcopy([self.projectOffice.runs[run],]), 
                                            observations=copy.deepcopy([self.projectOffice.scans[run],]))
            
            ## Observations
            for o,obs in enumerate(ses.scans):
                obs_start = mjdmpm2datetime(obs.mjd, obs.mpm)
                if isinstance(obs, DRX):
                    new_obs = sdf.DRX(obs.intent, obs.target, _UTC.localize(obs_start), obs.duration, 
                                      obs.ra, obs.dec, 
                                      obs.frequency1, obs.frequency2, obs.filter, 
                                      gain=obs.gain, MaxSNR=False, comments=obs.comments)
                elif isinstance(obs, Solar):
                    new_obs = sdf.Solar(obs.intent, obs.target, _UTC.localize(obs_start), obs.duration, 
                                        obs.frequency1, obs.frequency2, obs.filter, 
                                        gain=obs.gain, MaxSNR=False, comments=obs.comments)
                elif isinstance(obs, Jovian):
                    new_obs = sdf.Jovian(obs.intent, obs.target, _UTC.localize(obs_start), obs.duration, 
                                         obs.frequency1, obs.frequency2, obs.filter, 
                                         gain=obs.gain, MaxSNR=False, comments=obs.comments)
                else:
                    raise RuntimeError("This should never happen")
                session.append(new_obs)
                
                npc = len(obs.alt_phase_centers)
                for p,phase_center in enumerate(reversed(obs.alt_phase_centers)):
                    cid = npc - p
                    alt_t, alt_i, alt_r, alt_d = phase_center.target, phase_center.intent, phase_center.ra, phase_center.dec
                    try:
                        new_projoff.observations[0][o]
                    except IndexError:
                        new_projoff.observations[0][o] = None
                    if new_projoff.observations[0][o] is None:
                        new_projoff.observations[0][o] = ''
                    new_projoff.observations[0][o] = "alttarget%i:%s;;altintent%i:%s;;altra%i:%.9f;;altdec%i:%+.9f;;%s" % (cid, alt_t, cid, alt_i, cid, alt_r, cid, alt_d, new_projoff.observations[0][o])
                    
            ## Project
            project = sdf.Project(new_observer, "%s - %s (%i of %i)" % (self.name, station.id, i+1, len(ses.stations)), 
                                  copy.deepcopy(self.id), sessions=[session,], comments=copy.deepcopy(self.comments), 
                                  projectOffice=new_projoff)
            if project.projectOffice.sessions[0] is None:
               project.projectOffice.sessions[0] = ''
            project.projectOffice.sessions[0] = "corrchannels:%i;;corrinttime:%.3f;;corrbasis:%s;;origuser:%s;;origreturn:%s;;%s" % (ses.corr_channels, ses.corr_inttime, ses.corr_basis, ses.ucfuser, ses.dataReturnMethod.lower(), project.projectOffice.sessions[0])
            
            ## Save an increment the session ID
            sdfs.append(project)
            starting_session_id += 1
            
        # Done
        return sdfs


@cmp_to_total
class Run(object):
    """Class to hold all of the scans in an interferometer run."""
    
    def __init__(self, name, id, scans=None, dataReturnMethod='DRSU', comments=None, corr_channels=256, corr_inttime=1.0, corr_basis='linear', stations=getFullStations()):
        self.name = name
        self.id = int(id)
        if scans is None:
            self.scans = []
        else:
            if isinstance(scans, Scan):
                scans = [scans,]
            elif isinstance(scans, (tuple, list)):
                for i,obs in enumerate(scans):
                    if not isinstance(obs, Scan):
                        raise TypeError("Expected index %i of 'obsevations' to be an Scan" % i)
            else:
                raise TypeError("Expected 'scans' to be either a tuple or list of Scans of an Scan")
            self.scans = scans
        self.dataReturnMethod = dataReturnMethod
        self.ucfuser = None
        self.comments = comments
        
        self.corr_inttime = corr_inttime
        self.corr_channels = corr_channels
        self.corr_basis = corr_basis
        self.stations = stations
        
    def setStations(self, stations):
        """
        Update the stations used by the project for source computations.
        """
        
        if not isinstance(stations, (tuple, list)):
            raise TypeError('Expected a tuple of list of LWAStations')
        for i,station in enumerate(stations):
            if not isinstance(station, LWAStation):
                raise TypeError("Expected index %i to be an LWAStation" % i)
        self.stations = stations
        self.update()
        
    def append(self, newScan):
        """Add a new Scan to the list of scans."""
        
        if not isinstance(newScan, Scan):
            raise TypeError("Expected an Scan")
        self.scans.append(newScan)
        
    def setCorrelatorChannels(self, value):
        """Set the number of spectrometer channels to generate, 0 to disable."""
        
        self.corr_channels = int(value)
        
    def setCorrelatorInttime(self, value):
        """Set the number of spectrometer FFT integrations to use, 0 to disable."""
        
        self.corr_inttime = float(value)
        
    def setCorrelatorBasis(self, value):
        """Set the correlator output polarization basis."""
        
        if value.lower() not in ('linear', 'circular', 'stokes'):
            raise ValueError("Unknown polarization basis: %s" % value)
        self.corr_basis = value
        
    def setDataReturnMethod(self, method):
        """Set the data return method for the run.  Valid values are: UCF, DRSU, and 
        'USB Harddrives'."""
        
        if method not in ('UCF', 'DRSU', 'USB Harddrives'):
            raise ValueError("Unknown data return method: %s" % method)
            
        self.dataReturnMethod = method
        
    def setUCFUsername(self, username):
        """Set the username to use for UCF data copies."""
        
        self.ucfuser = username
        
    def update(self):
        """Update the various scans in the run."""
        
        for obs in self.scans:
            obs.update()
            
    def validate(self, verbose=False):
        """Examine all of the scans associated with the run to check
        for validity.  If everything is valid, return True.  Otherwise, return
        False."""
        
        failures = 0
        totalData = 0.0
        if len(self.stations) < 2:
            if verbose:
                print("[%i] Error: Need at least two stations to form an interferometer" % (os.getpid(),))
            failures += 1
        if self.corr_inttime < 0.1 or self.corr_inttime > 10.0:
            if verbose:
                print("[%i] Error: Invalid correlator integration time '%.3f s'" % (os.getpid(), self.corr_inttime))
            failures += 1
        if self.corr_channels < 16 or self.corr_channels > 32768 or self.corr_channels % 2:
            if verbose:
                print("[%i] Error: Invalid correlator channel count '%i'" % (os.getpid(), self.corr_channels))
            failures += 1
        if self.corr_basis.lower() not in (None, '', 'linear', 'circular', 'stokes'):
            if verbose:
                print("[%i] Error: Invalid correlator output polarization basis '%s'" % (os.getpid(), self.corr_basis))
            failures += 1
            
        scanCount = 1
        for obs in self.scans:
            if verbose:
                print("[%i] Validating scan %i" % (os.getpid(), scanCount))
                
            for station in self.stations:
                if not obs.validate(station=station, verbose=verbose):
                    failures += 1
                totalData += obs.dataVolume
                
            if scanCount > 1:
                if obs.filter != self.scans[scanCount-2].filter:
                    if verbose:
                        print("[%i] Error: Filter code changes at scan %i" % (os.getpid(), scanCount))
                    failures += 1
                    
            scanCount += 1
            
        # Make sure that the scans don't overlap
        sObs = self.scans
        
        for i in xrange(len(sObs)):
            maxOverlaps = 1
            overlaps = []
            nOverlaps = 0

            for j in xrange(len(sObs)):
                if verbose and i != j:
                    print("[%i] Checking for overlap between scans %i and %i" % (os.getpid(), i+1, j+1))

                cStart = int(sObs[j].mjd)*24*3600*1000 + int(sObs[j].mpm)
                cStop = cStart + int(sObs[j].dur)
                pStart = int(sObs[i].mjd)*24*3600*1000 + int(sObs[i].mpm)
                pStop = pStart + int(sObs[i].dur)

                if pStart >= cStart and pStart < cStop:
                    nOverlaps += 1
                    
                    if i != j:
                        overlaps.append(j)
            
            if nOverlaps > maxOverlaps:
                if verbose:
                    print("[%i] Error: Scan %i overlaps with %s" % (os.getpid(), i+1, ','.join(["%i" % (j+1) for j in overlaps])))
                failures += 1
            
        if totalData >= (len(self.stations)*_DRSUCapacityTB*1024**4):
            if verbose:
                print("[%i] Error: Total data volume for run exceeds per-station %i TB DRSU limit" % (os.getpid(), _DRSUCapacityTB,))
            failures += 1
        
        if failures == 0:
            return True
        else:
            return False
            
    def __cmp__(self, other):
        self.scans.sort()
        other.scans.sort()
        
        startSelf = self.scans[0].mjd + self.scans[0].mpm / (1000.0*3600.0*24.0)
        startOther = other.scans[0].mjd + other.scans[0].mpm / (1000.0*3600.0*24.0)
        if startSelf < startOther:
            return -1
        elif startSelf > startOther:
            return 1
        else:
            return 0


@cmp_to_total
class Scan(object):
    """
    Class to hold the specifics of a scans.  It currently
    handles TRK_RADEC, TRK_SOL, and TRK_JOV.
    """
    
    id = 1
    filterCodes = DRXFilters

    def __init__(self, target, intent, start, duration, mode, ra, dec, frequency1, frequency2, filter, gain=-1, comments=None):
        self.target = target
        self.intent = intent
        self.ra = float(ra) * (12.0/math.pi if type(ra).__name__ == 'Angle' else 1.0)
        self.dec = float(dec)* (180.0/math.pi if type(dec).__name__ == 'Angle' else 1.0)
        self.start = start
        if type(duration).__name__ == 'timedelta':
            # Make sure the number of microseconds agree with milliseconds
            us = int(round(duration.microseconds/1000.0))*1000
            duration = timedelta(days=duration.days, seconds=duration.seconds, microseconds=us)
        self.duration = str(duration)
        self.mode = mode
        self.beamDipole = None
        self.frequency1 = float(frequency1)
        self.frequency2 = float(frequency2)
        self.filter = int(filter)
        self.comments = comments
        
        self.mjd = None
        self.mpm = None
        self.dur = None
        self.freq1 = None
        self.freq2 = None
        self.dataVolume = None
        
        self.aspFlt = -1
        
        self.gain = int(gain)
        
        self.alt_phase_centers = []
        
        self.update()
        
    def __str__(self):
        """Return a nice string to describe the scan."""
        
        return "%s Obs. of '%s':\n Start %s\n Duration %s\n Filter: %i\n Frequency: %.3f; %.3f\n RA: %.3f\n Dec. %.3f\n" % (self.mode, self.name, self.start, self.duration, self.filter, self.frequency1, self.frequency2, self.ra, self.dec)
        
    def update(self):
        """Update the computed parameters from the string values."""
        
        # If we have a datetime instance, make sure we have an integer
        # number of milliseconds
        if type(self.start).__name__ == 'datetime':
            us = self.start.microsecond
            us = int(round(us/1000.0))*1000
            self.start = self.start.replace(microsecond=us)
        self.duration = str(self.duration)
        
        self.mjd = self.getMJD()
        self.mpm = self.getMPM()
        self.dur = self.getDuration()
        self.freq1 = self.getFrequency1()
        self.freq2 = self.getFrequency2()
        self.dataVolume = self.estimateBytes()
        
        # Update the associated alternate phase centers
        for phase_center in self.alt_phase_centers:
            phase_center.update()
            
    def setStart(self, start):
        """Set the scan start time."""
        
        self.start = start
        self.update()
        
    def getMJD(self):
        """Return the modified Julian Date corresponding to the date/time of the
        self.start string."""
        
        utc = sdf.parseTime(self.start)		## TODO:  We need to get the station informaiton here somehow
        utc = Time(utc, format=Time.FORMAT_PY_DATE)
        return int(utc.utc_mjd)

    def getMPM(self):
        """Return the number of milliseconds between the date/time specified in the
        self.start string and the previous UT midnight."""
        
        utc = sdf.parseTime(self.start)		## TODO:  We need to get the station informaiton here somehow
        utcMidnight = datetime(utc.year, utc.month, utc.day, 0, 0, 0, tzinfo=_UTC)
        diff = utc - utcMidnight
        return int(round((diff.seconds + diff.microseconds/1000000.0)*1000.0))

    def getDuration(self):
        """Parse the self.duration string with the format of HH:MM:SS.SSS to return the
        number of milliseconds in that period."""
        
        fields = self.duration.split(':')
        if len(fields) == 3:
            out = int(fields[0])*3600.0
            out += int(fields[1])*60.0
            out += float(fields[2])
        elif len(fields) == 2:
            out = int(fields[0])*60.0
            out += float(fields[1])
        else:
            out = float(fields[0])
            
        return int(round(out*1000.0))

    def setDuration(self, duration):
        """Set the scan duration."""
        
        if type(duration).__name__ == 'timedelta':
            # Make sure the number of microseconds agree with milliseconds
            us = int(round(duration.microseconds/1000.0))*1000
            duration = timedelta(days=duration.days, seconds=duration.seconds, microseconds=us)
        self.duration = str(duration)
        self.update()
        
    def getFrequency1(self):
        """Return the number of "tuning words" corresponding to the first frequency."""
        
        freq1 = freq2word(self.frequency1)
        self.frequency1 = word2freq(freq1)
        return freq1
        
    def setFrequency1(self, frequency1):
        """Set the frequency in Hz corresponding to tuning 1."""
        
        self.frequency1 = float(frequency1)
        self.update()
        
    def getFrequency2(self):
        """Return the number of "tuning words" corresponding to the second frequency."""
        
        freq2 = freq2word(self.frequency2)
        self.frequency2 = word2freq(freq2)
        return freq2
        
    def setFrequency2(self, frequency2):
        """Set the frequency in Hz correpsonding to tuning 2."""
        
        self.frequency2 = float(frequency2)
        self.update()
        
    def addAltPhaseCenter(self, target_or_apc, intent=None, ra=None, dec=None):
        """Add an alternate phase center to the scan."""
        
        if isinstance(target_or_apc, AlternatePhaseCenter):
            apc = target_or_apc
        else:
            apc = AlternatePhaseCenter(target_or_apc, intent, ra, dec)
        self.alt_phase_centers.append(apc)
        
    def estimateBytes(self):
        """Estimate the data volume for the specified type and duration of 
        scans.  For DRX:
        
            bytes = duration * sampleRate / 4096 * 4128 bytes * 2 tunings * 2 pols.
        """
        
        try:
            nFrames = self.getDuration()/1000.0 * self.filterCodes[self.filter] / 4096
        except KeyError:
            nFrames = 0
        nBytes = nFrames * DRXSize * 4
        return nBytes
        
    def getFixedBody(self):
        """Return an ephem.Body object corresponding to where the scan is 
        pointed.  None if the scan mode is TBN."""
        
        pnt = ephem.FixedBody()
        pnt.name = self.target
        pnt._ra = self.ra / 12.0 * math.pi
        pnt._dec = self.dec / 180.0 * math.pi
        pnt._epoch = '2000'
        return pnt
        
    def computeVisibility(self, station=lwa1):
        """Return the fractional visibility of the target during the scan 
        period."""
        
        lwa = station.getObserver()
        pnt = self.getFixedBody()
        
        vis = 0
        cnt = 0
        dt = 0.0
        while dt <= self.dur/1000.0:
            lwa.date = self.mjd + (self.mpm/1000.0 + dt)/3600/24.0 + MJD_OFFSET - DJD_OFFSET
            pnt.compute(lwa)
            
            cnt += 1
            if pnt.alt > 0:
                vis += 1
                
            dt += 300.0
        
        return float(vis)/float(cnt)
        
    def validate(self, station, verbose=False):
        """Evaluate the scan and return True if it is valid, False otherwise."""
        
        failures = 0
        # Basic - Intent, duration, frequency, and filter code values
        if self.intent.lower() not in ('fluxcal', 'phasecal', 'target'):
            if verbose:
                print("[%s] Error: Invalid scan intent '%s'" % (os.getpid(), self.intent))
            failures += 1
        if self.dur < 1:
            if verbose:
                print("[%i] Error: Specified a duration of length zero" % os.getpid())
            failures += 1
        if self.freq1 < 123809006 or self.freq1 > 2037918156:
            if verbose:
                print("[%i] Error: Specified frequency for tuning 1 is outside of DP tuning range" % os.getpid())
            failures += 1
        if (self.freq2 < 123809006 or self.freq2 > 2037918156) and self.freq2 != 0:
            if verbose:
                print("[%i] Error: Specified frequency for tuning 2 is outside of DP tuning range" % os.getpid())
            failures += 1
        if self.filter not in [1, 2, 3, 4, 5, 6]:
            if verbose:
                print("[%i] Error: Invalid filter code '%i'" % (os.getpid(), self.filter))
            failures += 1
            
        # Advanced - Target Visibility
        if self.ra < 0 or self.ra >= 24:
            if verbose:
                print("[%i] Error: Invalid value for RA '%.6f'" % (os.getpid(), self.ra))
            failures += 1
        if self.dec < -90 or self.dec > 90:
            if verbose:
                print("[%i] Error: Invalid value for dec. '%+.6f'" % (os.getpid(), self.dec))
            failures += 1
        if self.computeVisibility(station=station) < 1.0:
            if verbose:
                print("[%i] Error: Target is only above the horizon for %.1f%% of the scan for %s" % (os.getpid(), self.computeVisibility(station=station)*100.0, station.id))
            failures += 1
            
        # Advanced - alternate phase centers
        if len(self.alt_phase_centers) > 5:
            if verbose:
                print("[%i] Error: too many alternate phase centers defined" % os.getpid())
            failures += 1
        for j,phase_center in enumerate(self.alt_phase_centers):
            if not phase_center.validate(station, verbose=verbose):
                if verbose:
                    print("[%i] Error: invalid alternate phase center %i" % (os.getpid(), j+1))
                failures += 1
                
            ## Closeness to pointing center
            pnt = self.getFixedBody()
            alt_pnt = phase_center.getFixedBody()
            
            lwa = station.getObserver()
            lwa.date = self.mjd + (self.mpm/1000.0 + self.dur/1000.0/2.0)/3600/24.0 + MJD_OFFSET - DJD_OFFSET
            pnt.compute(lwa)
            alt_pnt.compute(lwa)
            
            alt_sep = ephem.separation(pnt, alt_pnt) * 180/math.pi
            
            beam = 2.0*74e6/max([self.frequency1, self.frequency2])
            if alt_sep > beam/2.0:
                if verbose:
                    print("[%i] Error: alternate phase center %i is %.1f degrees from pointing center" % (os.getpid(), j+1, alt_sep))
                failures += 1
                
        # Advanced - Data Volume
        if self.dataVolume >= (_DRSUCapacityTB*1024**4):
            if verbose:
                print("[%i] Error: Data volume exceeds %i TB DRSU limit" % (os.getpid(), _DRSUCapacityTB))
            failures += 1
            
        # Any failures indicates a bad scan
        if failures == 0:
            return True
        else:
            return False
            
    def __cmp__(self, other):
        startSelf = self.mjd + self.mpm / (1000.0*3600.0*24.0)
        startOther = other.mjd + other.mpm / (1000.0*3600.0*24.0)
        if startSelf < startOther:
            return -1
        elif startSelf > startOther:
            return 1
        else:
            return 0


class DRX(Scan):
    """
    Required Arguments:
     * scan target
     * scan intent
     * scan start date/time (UTC YYYY/MM/DD HH:MM:SS.SSS string or timezone-
       aware datetime instance)
     * scan duration (HH:MM:SS.SSS string or timedelta instance)
     * scan RA in hours, J2000.0 or ephem.hours instance
     * scan Dec in degrees, J2000.0 or ephem.hours instance
     * scan tuning frequency 1 (Hz)
     * scan tuning frequency 1 (Hz)
     * integer filter code
    
    Optional Keywords:
     * comments - comments about the scan
    """
    
    def __init__(self, target, intent, start, duration, ra, dec, frequency1, frequency2, filter, gain=-1, comments=None):
        Scan.__init__(self, target, intent, start, duration, 'TRK_RADEC', ra, dec, frequency1, frequency2, filter, gain=gain, comments=comments)
        
    def setRA(self, ra):
        """Set the pointing RA."""
        
        self.ra = float(ra) * (12.0/math.pi if type(ra).__name__ == 'Angle' else 1.0)
        
    def setDec(self, dec):
        """Set the pointing Dec."""
        
        self.dec = float(dec)* (180.0/math.pi if type(dec).__name__ == 'Angle' else 1.0)
        
        
class Solar(Scan):
    """Sub-class of DRX specifically for Solar DRX scans.   It features a
    reduced number of parameters needed to setup the scan.
    
    Required Arguments:
     * scan target
     * scan intent
     * scan start date/time (UTC YYYY/MM/DD HH:MM:SS.SSS string or timezone-
       aware datetime instance)
     * scan duration (HH:MM:SS.SSS string or timedelta instance)
     * scan tuning frequency 1 (Hz)
     * scan tuning frequency 1 (Hz)
     * integer filter code
    
    Optional Keywords:
     * comments - comments about the scan
    """
    
    def __init__(self, target, intent, start, duration, frequency1, frequency2, filter, gain=-1, comments=None):
        Scan.__init__(self, target, intent, start, duration, 'TRK_SOL', 0.0, 0.0, frequency1, frequency2, filter, gain=gain, comments=comments)
        
    def getFixedBody(self):
        """Return an ephem.Body object corresponding to where the scan is 
        pointed.  None if the scan mode is TBN."""
        
        return ephem.Sun()


class Jovian(Scan):
    """Sub-class of DRX specifically for Jovian DRX scans.   It features a
    reduced number of parameters needed to setup the scan.
    
    Required Arguments:
     * scan target
     * scan intent
     * scan start date/time (UTC YYYY/MM/DD HH:MM:SS.SSS string or timezone-
       aware datetime instance)
     * scan duration (HH:MM:SS.SSS string or timedelta instance)
     * scan tuning frequency 1 (Hz)
     * scan tuning frequency 1 (Hz)
     * integer filter code
    
    Optional Keywords:
     * comments - comments about the scan
    """
    
    def __init__(self, target, intent, start, duration, frequency1, frequency2, filter, gain=-1, comments=None):
        Scan.__init__(self, target, intent, start, duration, 'TRK_JOV', 0.0, 0.0, frequency1, frequency2, filter, gain=gain, comments=comments)

    def getFixedBody(self):
        """Return an ephem.Body object corresponding to where the scan is 
        pointed.  None if the scan mode is TBN."""
        
        return ephem.Jupiter()


class AlternatePhaseCenter(object):
    """Class to hold an alternate phase center for a scan."""
    
    def __init__(self, target, intent, ra, dec):
        self.target = target
        self.intent = intent
        self.ra = float(ra) * (12.0/math.pi if type(ra).__name__ == 'Angle' else 1.0)
        self.dec = float(dec)* (180.0/math.pi if type(dec).__name__ == 'Angle' else 1.0)
        
    def setRA(self, ra):
        """Set the pointing RA."""
        
        self.ra = float(ra) * (12.0/math.pi if type(ra).__name__ == 'Angle' else 1.0)
        
    def setDec(self, dec):
        """Set the pointing Dec."""
        
        self.dec = float(dec)* (180.0/math.pi if type(dec).__name__ == 'Angle' else 1.0)
        
    def update(self):
        """Update the computed parameters from the string values."""
        
        pass
        
    def getFixedBody(self):
        """Return an ephem.Body object corresponding to where the scan is 
        pointed.  None if the scan mode is TBN."""
        
        pnt = ephem.FixedBody()
        pnt.name = self.target
        pnt._ra = self.ra / 12.0 * math.pi
        pnt._dec = self.dec / 180.0 * math.pi
        pnt._epoch = '2000'
        return pnt
    
    def validate(self, station, verbose=False):
        """Basic validation of the pointing, that's it."""
        
        failures = 0
        
        ## Intent
        if self.intent.lower() not in ('fluxcal', 'phasecal', 'target'):
            if verbose:
                print("[%s] Error: Invalid alternate phase center intent '%s'" % (os.getpid(), self.intent))
            failures += 1
            
        ## Pointing
        if self.ra < 0 or self.ra >= 24:
            if verbose:
                print("[%i] Error: Invalid alternate phase center value for RA '%.6f'" % (os.getpid(), self.ra))
            failures += 1
        if self.dec < -90 or self.dec > 90:
            if verbose:
                print("[%i] Error: Invalid alternate phase center value for dec. '%+.6f'" % (os.getpid(), self.dec))
            failures += 1
            
        # Any failures indicates a bad scan
        if failures == 0:
            return True
        else:
            return False


def __parseCreateScanObject(obsTemp, altTemps=[], verbose=False):
    """Given a obsTemp dictionary of scan parameters, return a complete Scan object 
    corresponding to those values."""
    
    # If the scan ID is 0, do nothing.
    if obsTemp['id'] == 0:
        return None
        
    # Create a time string for the start time in UTC.  This is a little tricky 
    # because of the rounding to the nearest millisecond which has to be done
    # to the datetime object.
    start = Time(obsTemp['mjd'] + obsTemp['mpm'] / 1000.0 / 3600.0 / 24.0, format='MJD').utc_py_date
    start += timedelta(microseconds=(int(round(start.microsecond/1000.0)*1000.0)-start.microsecond))
    utcString = start.strftime("UTC %Y %m %d %H:%M:%S.%f")
    utcString = utcString[:-3]
    
    # Build up a string representing the scan duration.
    try:
        dur = obsTemp['duration']
        dur = float(dur) / 1000.0
        durString = '%02i:%02i:%06.3f' % (dur/3600.0, (dur%3600.0)/60.0, dur%60.0)
    except:
        pass
        
    # Convert the frequencies from "tuning words" to Hz
    f1 = word2freq(obsTemp['freq1'])
    f2 = word2freq(obsTemp['freq2'])
    
    # Get the mode and run through the various cases
    mode = obsTemp['mode']
    if verbose:
        print("[%i] Scan %i is mode %s" % (os.getpid(), obsTemp['id'], mode))
        
    if mode == 'TRK_RADEC':
        obsOut = DRX(obsTemp['target'], obsTemp['intent'], utcString, durString, obsTemp['ra'], obsTemp['dec'], f1, f2, obsTemp['filter'], gain=obsTemp['gain'], comments=obsTemp['comments'])
    elif mode == 'TRK_SOL':
        obsOut = Solar(obsTemp['target'], obsTemp['intent'], utcString, durString, f1, f2, obsTemp['filter'], gain=obsTemp['gain'], comments=obsTemp['comments'])
    elif mode == 'TRK_JOV':
        obsOut = Jovian(obsTemp['target'], obsTemp['intent'], utcString, durString, f1, f2, obsTemp['filter'], gain=obsTemp['gain'], comments=obsTemp['comments'])
    else:
        raise RuntimeError("Invalid mode encountered: %s" % mode)
        
    # Add in the alternate phase centers
    for altTemp in altTemps:
        obsOut.addAltPhaseCenter(altTemp['target'], altTemp['intent'], altTemp['ra'], altTemp['dec'])
        
    # Set the ASP/FEE values
    obsOut.aspFlt = copy.deepcopy(obsTemp['aspFlt'])
    
    # Force the scan to be updated
    obsOut.update()
    
    # Return the newly created Scan object
    return obsOut


def parseIDF(filename, verbose=False):
    """
    Given a filename, read the file's contents into the IDF instance and return
    that instance.
    """
    
    # Open the file
    fh = open(filename, 'r')
    
    # Create the keyword regular expression to deal with various indicies included 
    # in the keywords
    kwdRE = re.compile(r'(?P<keyword>[A-Z_0-9\+]+)(\[(?P<id1>[0-9]+?)\])?(\[(?P<id2>[0-9]+?)\])?(\[(?P<id3>[0-9]+?)\])?(\[(?P<id4>[0-9]+?)\])?')
    
    # Create the metatag regular expression to deal with spectrometer mode settings
    metaRE = re.compile(r'\{.*\}')
    
    # Create empty objects to get things started.  Values will get filled in as they
    # are found in the file
    po = ProjectOffice()
    observer = Observer('observer_name', 0)
    project = Project(observer, 'project_name', 'project_id', projectOffice=po)
    run = Run('run_name', 0, scans=[])
    project.runs = [run,]
    project.projectOffice.runs = []
    project.projectOffice.scans = [[],]
    
    # Loop over the file
    obsTemp = {'id': 0, 'target': '', 'intent': '', 'ra': 0.0, 'dec': 0.0, 'start': '', 'duration': '', 'mode': '', 
               'freq1': 0, 'freq2': 0, 'filter': 0, 'comments': None, 'gain': -1, 
               'aspFlt': -1}
    altTemp = {'id': 0, 'target': '', 'intent': '', 'ra': 0.0, 'dec': 0.0}
    altTemps = []
    
    for line in fh:
        # Trim off the newline character and skip blank lines
        line = line.replace('\n', '')
        if len(line) == 0 or line.isspace():
            continue
            
        # Split into a keyword, value pair and run it through the regular expression
        # to deal with any indicies present
        try:
            keywordSection, value = line.split(None, 1)
        except:
            continue
            
        mtch = kwdRE.match(keywordSection)
        keyword = mtch.group('keyword')
        
        ids = [-1, -1, -1, -1]
        for i in xrange(4):
            try:
                ids[i] = int(mtch.group('id%i' % (i+1)))
            except TypeError:
                pass
                
        # Skip over the observer comment lines (denoted by a plus sign at the end) 
        # of the keyword
        if keyword[-1] == '+':
            continue
            
        # Observer Info
        if keyword == 'PI_ID':
            project.observer.id = int(value)
            continue
        if keyword == 'PI_NAME':
            project.observer.name = value
            project.observer.splitName()
            continue
            
        # Project/Proposal Info
        if keyword == 'PROJECT_ID':
            project.id = value
            continue
        if keyword == 'PROJECT_TITLE':
            project.name = value
            continue
        if keyword == 'PROJECT_REMPI':
            project.comments = value
            continue
        if keyword == 'PROJECT_REMPO':
            project.projectOffice.project = value
            continue
            
        # Run Info
        if keyword == 'RUN_ID':
            project.runs[0].id = int(value)
            continue
        if keyword == 'RUN_TITLE':
            project.runs[0].name = value
            continue
        if keyword == 'RUN_STATIONS':
            project.runs[0].stations = []
            possible = getFullStations()
            for field in value.split(','):
                field = field.strip().rstrip()
                if field.lower() == 'all':
                    run.stations = possible
                    break
                else:
                    for station in possible:
                        if station.id == field:
                            run.stations.append(station)
                            break
        if keyword == 'RUN_CHANNELS':
            project.runs[0].corr_channels = int(value)
            continue
        if keyword == 'RUN_INTTIME':
            project.runs[0].corr_inttime = float(value)
            continue
        if keyword == 'RUN_BASIS':
            project.runs[0].corr_basis = value
            continue
        if keyword == 'RUN_REMPI':
            mtch = sdf._usernameRE.search(value)
            if mtch is not None:
                project.runs[0].ucfuser = mtch.group('username')
                if mtch.group('subdir') is not None:
                    project.runs[0].ucfuser = os.path.join(project.runs[0].ucfuser, mtch.group('subdir'))
            project.runs[0].comments = sdf._usernameRE.sub('', value)
            continue
        if keyword == 'RUN_REMPO':
            project.projectOffice.runs.append(None)
            parts = value.split(';;', 1)
            first = parts[0]
            try:
                second = parts[1]
            except IndexError:
                second = ''
                
            if first[:31] == 'Requested data return method is':
                # Catch for project office comments that are data return related
                project.runs[0].dataReturnMethod = first[32:]
                project.projectOffice.runs[0] = second
            else:
                # Catch for standard (not data related) project office comments
                project.projectOffice.runs[0] = value
            continue
            
        # Scan Info
        if keyword == 'SCAN_ID':
            if obsTemp['id'] != 0:
                project.runs[0].scans.append( __parseCreateScanObject(obsTemp, altTemps=altTemps, verbose=verbose) )
                altTemp = {'id': 0, 'target': '', 'intent': '', 'ra': 0.0, 'dec': 0.0}
                altTemps = []
            obsTemp['id'] = int(value)
            project.projectOffice.scans[0].append( None )
            
            if verbose:
                print("[%i] Started scan %i" % (os.getpid(), int(value)))
                
            continue
        if keyword == 'SCAN_TARGET':
            obsTemp['target'] = value
            continue
        if keyword == 'SCAN_INTENT':
            obsTemp['intent'] = value
            continue
        if keyword == 'SCAN_REMPI':
            obsTemp['comments'] = value
            continue
        if keyword == 'SCAN_REMPO':
            project.projectOffice.scans[0][-1] = value
            continue
        if keyword == 'SCAN_START_MJD':
            obsTemp['mjd'] = int(value)
            continue
        if keyword == 'SCAN_START_MPM':
            obsTemp['mpm'] = int(value)
            continue
        if keyword == 'SCAN_DUR':
            obsTemp['duration'] = int(value)
            continue
        if keyword == 'SCAN_MODE':
            obsTemp['mode'] = value
            continue
        if keyword == 'SCAN_RA':
            obsTemp['ra'] = float(value)
            continue
        if keyword == 'SCAN_DEC':
            obsTemp['dec'] = float(value)
            continue
        if keyword == 'SCAN_FREQ1':
            obsTemp['freq1'] = int(value)
            continue
        if keyword == 'SCAN_FREQ2':
            obsTemp['freq2'] = int(value)
            continue
        if keyword == 'SCAN_BW':
            obsTemp['filter'] = int(value)
            continue
        
        # Alternate phase centers
        if keyword == 'SCAN_ALT_TARGET':
            if len(altTemps) == 0:
                altTemps.append( copy.deepcopy(altTemp) )
                altTemps[-1]['id'] = ids[0]
                altTemps[-1]['target'] = value
            else:
                if altTemps[-1]['id'] != ids[0]:
                    altTemps.append( copy.deepcopy(altTemps[-1]) )
                    altTemps[-1]['id'] = ids[0]
                altTemps[-1]['target'] = value
            continue
        if keyword == 'SCAN_ALT_INTENT':
            if len(altTemps) == 0:
                altTemps.append( copy.deepcopy(altTemp) )
                altTemps[-1]['id'] = ids[0]
                altTemps[-1]['intent'] = value
            else:
                if altTemps[-1]['id'] != ids[0]:
                    altTemps.append( copy.deepcopy(altTemps[-1]) )
                    altTemps[-1]['id'] = ids[0]
                altTemps[-1]['intent'] = value
            continue
        if keyword == 'SCAN_ALT_RA':
            if len(altTemps) == 0:
                altTemps.append( copy.deepcopy(altTemp) )
                altTemps[-1]['id'] = ids[0]
                altTemps[-1]['ra'] = float(value)
            else:
                if altTemps[-1]['id'] != ids[0]:
                    altTemps.append( copy.deepcopy(altTemps[-1]) )
                    altTemps[-1]['id'] = ids[0]
                altTemps[-1]['ra'] = float(value)
            continue
        if keyword == 'SCAN_ALT_DEC':
            if len(altTemps) == 0:
                altTemps.append( copy.deepcopy(altTemp) )
                altTemps[-1]['id'] = ids[0]
                altTemps[-1]['dec'] = float(value)
            else:
                if altTemps[-1]['id'] != ids[0]:
                    altTemps.append( copy.deepcopy(altTemps[-1]) )
                    altTemps[-1]['id'] = ids[0]
                altTemps[-1]['dec'] = float(value)
            continue
            
        # Run wide settings at the end of the scans
        if keyword == 'SCAN_ASP_FLT':
            obsTemp['aspFlt'] = int(value)
            continue
        if keyword == 'SCAN_DRX_GAIN':
            obsTemp['gain'] = int(value)
            continue
            
        # Keywords that might indicate this is a SDF
        if keyword in ('SESSION_ID', 'SESSION_DRX_BEAM'):
            raise RuntimeError("Invalid keyword encountered: %s" % keyword)
            
    # Create the final scan
    if obsTemp['id'] != 0:
        project.runs[0].scans.append( __parseCreateScanObject(obsTemp, altTemps=altTemps, verbose=verbose) )
        
    # Close the file
    fh.close()
    
    # Return the project
    return project


def getScanStartStop(obs):
    """
    Given a scan, get the start and stop times (returned as a two-
    element tuple of UTC datetime instances).
    """
    
    # UNIX timestamp for the start
    tStart = utcjd_to_unix(obs.mjd + MJD_OFFSET)
    tStart += obs.mpm / 1000.0
    
    # UNIX timestamp for the stop
    tStop = tStart +  obs.dur / 1000.0
    
    # Conversion to a timezone-aware datetime instance
    tStart = _UTC.localize( datetime.utcfromtimestamp(tStart) )
    tStop  = _UTC.localize( datetime.utcfromtimestamp(tStop ) )
    
    # Make sure we have an integer number of milliseconds
    ## Start
    us = tStart.microsecond
    us = int(round(us/1000.0))*1000
    tStart = tStart.replace(microsecond=us)
    ## Stop
    us = tStop.microsecond
    us = int(round(us/1000.0))*1000
    tStop = tStop.replace(microsecond=us)
    
    # Return
    return tStart, tStop


def isValid(filename, verbose=False):
    """
    Given a filename, see if it is valid IDF file or not.
    """
    
    passes = 0
    failures = 0
    try:
        proj = parseIDF(filename)
        passes += 1
        if verbose:
            print("Parser - OK")
            
        valid = proj.validate()
        if valid:
            passes += 1
            if verbose:
                print("Validator - OK")
        else:
            failures += 1
            if verbose:
                print("Validator - FAILED")
                
    except IOError as e:
        raise e
    except:
        failures += 1
        if verbose:
            print("Parser - FAILED")
            
    if verbose:
        print("---")
        print("%i passed / %i failed" % (passes, failures))
        
    return False if failures else True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import math
import time
import argparse
from datetime import datetime

from lsl.reader import drx, errors
from lsl.common.progress import ProgressBar
from lsl.misc import parser as aph


def fileSplitFunction(fhIn, fhOut, nCaptures, nBeampols):
    pb = ProgressBar(max=nCaptures)
    
    for c in xrange(int(nCaptures)):
        for i in xrange(nBeampols):
            cFrame = fhIn.read(drx.FrameSize)
            fhOut.write(cFrame)
            
        pb.inc(amount=1)
        if c != 0 and c % 100 == 0:
            sys.stdout.write(pb.show()+'\r')
            sys.stdout.flush()
            
    sys.stdout.write(pb.show()+'\r')
    sys.stdout.write('\n')
    sys.stdout.flush()


def main(args):
    filename = args.filename

    sizeB = os.path.getsize(filename)

    # Open the file and get some basic info about the data contained
    fh = open(filename, 'rb')
    nFramesFile = sizeB / drx.FrameSize

    while True:
        try:
            junkFrame = drx.readFrame(fh)
            try:
                srate = junkFrame.getSampleRate()
                t0 = junkFrame.getTime()
                break
            except ZeroDivisionError:
                pass
        except errors.syncError:
            fh.seek(-drx.FrameSize+1, 1)
            
    fh.seek(-drx.FrameSize, 1)
    
    beams = drx.getBeamCount(fh)
    tunepols = drx.getFramesPerObs(fh)
    tunepol = tunepols[0] + tunepols[1] + tunepols[2] + tunepols[3]
    beampols = tunepol

    # Offset in frames for beampols beam/tuning/pol. sets
    offset = int(round(args.offset * srate / 4096 * beampols))
    offset = int(1.0 * offset / beampols) * beampols
    fh.seek(offset*drx.FrameSize, 1)
    
    # Iterate on the offsets until we reach the right point in the file.  This
    # is needed to deal with files that start with only one tuning and/or a 
    # different sample rate.  
    while True:
        ## Figure out where in the file we are and what the current tuning/sample 
        ## rate is
        junkFrame = drx.readFrame(fh)
        srate = junkFrame.getSampleRate()
        t1 = junkFrame.getTime()
        tunepols = drx.getFramesPerObs(fh)
        tunepol = tunepols[0] + tunepols[1] + tunepols[2] + tunepols[3]
        beampols = tunepol
        fh.seek(-drx.FrameSize, 1)
        
        ## See how far off the current frame is from the target
        tDiff = t1 - (t0 + args.offset)
        
        ## Half that to come up with a new seek parameter
        tCorr = -tDiff / 2.0
        cOffset = int(tCorr * srate / 4096 * beampols)
        cOffset = int(1.0 * cOffset / beampols) * beampols
        offset += cOffset
        
        ## If the offset is zero, we are done.  Otherwise, apply the offset
        ## and check the location in the file again/
        if cOffset is 0:
            break
        fh.seek(cOffset*drx.FrameSize, 1)
    
    # Update the offset actually used
    args.offset = t1 - t0
    
    nCaptures = nFramesFile/beampols

    print "Filename:     %s" % filename
    print "Size:         %.1f MB" % (float(sizeB)/1024/1024)
    print "Captures:     %i (%.2f seconds)" % (nCaptures, nCaptures*4096/srate)
    print "Tuning/Pols.: %i " % tunepol
    print "Sample Rate: %.2f MHz" % (srate/1e6)
    print "==="

    if args.count > 0:
        nCaptures = args.count * srate / 4096
    else:
        nCaptures -= args.offset * srate / 4096
        
        args.count = nCaptures * 4096 / srate

    print "Seconds to Skip:  %.2f (%i captures)" % (args.offset, offset/beampols)
    print "Seconds to Split: %.2f (%i captures)" % (args.count, nCaptures)

    # Make sure that the first frame in the file is the first frame of a capture 
    # (tuning 1, pol 0).  If not, read in as many frames as necessary to get to 
    # the beginning of a complete capture.
    beam, tune, pol = junkFrame.parseID()

    skip = 0
    while (2*(tune-1)+pol) != 0:
        frame = drx.readFrame(fh)
        beam, tune, pol = frame.parseID()
        skip += 1

    nFramesRemaining = (sizeB - fh.tell()) / drx.FrameSize
    nRecursions = int(nFramesRemaining / (nCaptures*beampols))
    if not args.recursive:
        nRecursions = 1
        
    scale = int(math.log10(nRecursions)) + 1
    ifString = "Working on #%%%ii of %i (%%s)" % (scale, nRecursions)
    
    for r in xrange(nRecursions):
        if args.date:
            filePos = fh.tell()
            junkFrame = drx.readFrame(fh)
            fh.seek(filePos)

            dt = datetime.utcfromtimestamp(junkFrame.getTime())
            captFilename = "%s_%s.dat" % (os.path.splitext(os.path.basename(filename))[0], dt.isoformat())
        else:
            captFilename = "%s_s%04i_p%%0%ii.dat" % (os.path.splitext(os.path.basename(filename))[0], args.count, scale)
            captFilename = captFilename % r
            if not args.recursive:
                captFilename = "%s_s%04i.dat" % (os.path.splitext(os.path.basename(filename))[0], args.count)
            
        print ifString % (r+1, captFilename)
        
        t0 = time.time()
        fhOut = open(captFilename, 'wb')
        fileSplitFunction(fh, fhOut, nCaptures, beampols)
        fhOut.close()
        t1 = time.time()
        print "  Copied %i bytes in %.3f s (%.3f MB/s)" % (os.path.getsize(captFilename), t1-t0, os.path.getsize(captFilename)/1024.0**2/(t1-t0))
        
    fh.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='split a DRX file into several files', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('filename', type=str, 
                        help='filename to split')
    parser.add_argument('-c', '--count', type=aph.positive_float, default=10.0, 
                        help='number of seconds to keep')
    parser.add_argument('-o', '--offset', type=aph.positive_or_zero_float, default=0.0, 
                        help='number of seconds to skip before splitting')
    parser.add_argument('-d', '--date', action='store_true', 
                        help='label the split files with a date rather than a sequence number')
    parser.add_argument('-r', '--recursive', action='store_true', 
                        help='recursively split the file')
    args = parser.parse_args()
    main(args)
    



# -*- coding: utf-8 -*

"""
Module that stores various useful constants in one convenient location.
The constants defined in this file are:

c
    the speed of light in m/s

kB
    Boltzmann's constant in Jy m^2 / K

deg2rad
    the conversion factor for degrees to radians

tpi
    :math:`2 \\pi \\sqrt{-1}`
"""

import math

__version__ = '0.2'
__revision__ = '$Rev: 2567 $'
__all__ = ['c', 'kB', 'deg2rad', 'tpi', '__version__', '__revision__', '__all__']


import warnings
warnings.warn("lsl.common.constants will be removed in LSL version 1.3.0", DeprecationWarning)


c = 2.9979245800e8			# speed of light in m/s
kB = 1380.6488				# Boltzmann's constant in Jy m^2 / K
deg2rad = math.pi / 180.0	# degrees to radians conversion
tpi = 2j*math.pi			# two pi i



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example script for reading in TBW data and writing it to a TS-FITS file."""

import os
import sys
import time
from lsl.reader import tbw
from lsl.reader import errors
from lsl.writer import tsfits

import matplotlib.pyplot as plt


def main(args):
    # Determine the number of samples in the specified file
    nSamples = os.path.getsize(args[0]) / tbw.FrameSize
    print "Samples in file: ", nSamples
    fh = open(args[0], "rb", buffering=tbw.FrameSize)

    # Make sure that the data is TBW and determine the data length
    test = tbw.readFrame(fh)
    print "TBW Data:  %s" % test.header.isTBW()
    if not test.header.isTBW():
        raise errors.notTBWError()
    print "Data Length: %i bits" % test.getDataBits()
    if test.header.getDataBits() == 12:
        nData = 400
    else:
        nData = 1200
    fh.seek(-tbw.FrameSize, 1)
    
    # Due to the size of the FITS files being generated, the number of frames that 
    # can be read in is limited to 300,000, or 30,000 frames for 10 stands.  Getting
    # around this limit will require finding out how to do on-the-fly FITS binary 
    # table resizing.  
    nSamples = 900000
    
    tStart = time.time()
    
    # Create a new FITS file with the name 'tbw.fits'
    fitsFile = tsfits.TBW('tbw-tsfits-test.fits')

    # Read in the data and add it to the FITS file created above
    count = {}
    syncCount = 0
    masterCount = 0
    for i in range(nSamples):
        # Read in the next frame and anticipate any problems that could occur
        try:
            cFrame = tbw.readFrame(fh, Verbose=False)
        except errors.eofError:
            break
        except errors.syncError:
            #print "WARNING: Mark 5C sync error on frame #%i" % (int(fh.tell())/TBWFrameSize-1)
            syncCount = syncCount + 1
            continue
            
        stand = cFrame.header.parseID()
        if cFrame.header.frameCount % 10000 == 0:
            print "%2i  %14i  %6.3f  %5i  %5i" % (stand, cFrame.data.timeTag, cFrame.getTime(), cFrame.header.frameCount, cFrame.header.secondsCount)
        if stand not in count.keys():
            count[stand] = 0
            
        fitsFile.addStandData(cFrame)
        
        count[stand] = count[stand] + 1
        masterCount = masterCount + 1
        
    tEnd = time.time()
    print 'Read %i frames in %0.3f s (%0.1f frames/s)' % (masterCount, (tEnd-tStart), masterCount/(tEnd-tStart))
    
    fh.close()
    fitsFile.close()
    fitsFile.info()
    
    # Summary information about the file that was just read in
    print "Summary:"
    for stand in sorted(count.keys()):
        print "Stand: %2i, Frames: %5i" % (stand, count[stand])
    print "Sync Errors: %5i" % syncCount


if __name__ == "__main__":
    main(sys.argv[1:])
